import datetime
import os
from pathlib import Path
import sys
import asyncio
import subprocess
import threading
import json
from time import perf_counter


import modal


app = modal.App("openai_gpt_oss")
IMAGE_GPU = os.getenv("IMAGE_GPU", None)
BACKEND = os.getenv("BACKEND", "")
APP_NAME = os.getenv("APP_NAME", "")
TP = os.getenv("TP", "1")

img = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
    modal.Image.from_registry(
        "nvcr.io/nvidia/cuda:12.9.1-cudnn-devel-ubuntu22.04",
        add_python="3.12",
    )
    .apt_install("git", "git-lfs")
    .run_commands(
        "uv pip install --system --pre --no-cache vllm==0.10.1+gptoss "
        "--extra-index-url https://wheels.vllm.ai/gpt-oss/ "
        "--extra-index-url https://download.pytorch.org/whl/nightly/cu128 "
        "--index-strategy unsafe-best-match",
    )
    .pip_install("openai-harmony")
    .env(
        {
            "PYTORCH_CUDA_ALLOC_CONF": "expandable_segments:True",
            # "TQDM_DISABLE": "1",
            "LLM_MODEL": os.getenv("LLM_MODEL", "openai/gpt-oss-20b"),
            "VLLM_USE_V1": "1",
            "TP": TP,
            "VLLM_WORKER_MULTIPROC_METHOD": "spawn",
            "TORCH_CUDA_ARCH_LIST": "8.0 8.9 9.0+PTX",
            "VLLM_ATTENTION_BACKEND": "TRITON_ATTN_VLLM_V1",
        }
    )
    .run_commands("git clone https://github.com/weedge/gpt-oss.git")
)
if BACKEND == "flashinfer":
    img = img.pip_install(
        f"flashinfer-python",
        extra_index_url="https://wheels.vllm.ai/gpt-oss/",
    )


HF_MODEL_DIR = "/root/.achatbot/models"
hf_model_vol = modal.Volume.from_name("models", create_if_missing=True)
VLLM_CACHE_DIR = "/root/.cache/vllm"
vllm_cache_vol = modal.Volume.from_name("vllm-cache", create_if_missing=True)


with img.imports():
    import termcolor
    import torch
    from vllm import LLM, LLMEngine, EngineArgs, SamplingParams, TokensPrompt
    from openai_harmony import (
        HarmonyEncodingName,
        load_harmony_encoding,
        Conversation,
        Message,
        Role,
        SystemContent,
        DeveloperContent,
        ReasoningEffort,
        StreamableParser,
        StreamState,
        TextContent,
        ToolDescription,
        Author,
    )

    sys.path.insert(0, "/gpt-oss")
    from gpt_oss.tools import apply_patch
    from gpt_oss.tools.simple_browser import SimpleBrowserTool
    from gpt_oss.tools.simple_browser.backend import ExaBackend
    from gpt_oss.tools.python_docker.docker_tool import PythonTool

    REASONING_EFFORT = {
        "high": ReasoningEffort.HIGH,
        "medium": ReasoningEffort.MEDIUM,
        "low": ReasoningEffort.LOW,
    }

    MODEL_PATH = os.getenv("LLM_MODEL", "openai/gpt-oss-20b")
    model_path = os.path.join(HF_MODEL_DIR, MODEL_PATH)
    model_name = MODEL_PATH.split("/")[-1]

    class TokenGenerator:
        def __init__(
            self,
            model_path: str,
            tensor_parallel_size: int = 1,
            gpu_memory_utilization: float = 0.7,
        ):
            args = EngineArgs(
                model=model_path,
                tensor_parallel_size=tensor_parallel_size,
                gpu_memory_utilization=gpu_memory_utilization,
                # quantization="mxfp4",
            )
            self.engine = LLMEngine.from_engine_args(args)
            self.request_id = 0

        def generate(
            self,
            prompt_tokens: list[int],
            stop_tokens: list[int] | None = None,
            temperature: float = 1.0,
            top_p: float = 1.0,
            max_tokens: int = 0,
            return_logprobs: bool = False,
        ):
            if max_tokens == 0:
                max_tokens = None
            request_id = str(self.request_id)
            self.request_id += 1
            sampling_params = SamplingParams(
                temperature=temperature,
                top_p=top_p,
                max_tokens=max_tokens,
                stop_token_ids=stop_tokens,
                logprobs=0 if return_logprobs else None,
            )
            prompt = TokensPrompt(prompt_token_ids=prompt_tokens)
            self.engine.add_request(request_id, prompt, sampling_params)
            last_token_id = []
            while self.engine.has_unfinished_requests():
                step_outputs = self.engine.step()
                output = step_outputs[0].outputs[0]
                token_ids = output.token_ids
                logprobs_list = output.logprobs if hasattr(output, "logprobs") else None
                new_token_ids = token_ids[len(last_token_id) :]
                new_logprobs = (
                    logprobs_list[len(last_token_id) :]
                    if logprobs_list is not None
                    else [None] * len(new_token_ids)
                )
                for token_id, logprobs in zip(new_token_ids, new_logprobs):
                    last_token_id.append(token_id)
                    if return_logprobs:
                        logprob_val = None
                        if logprobs is not None and token_id in logprobs:
                            logprob_val = logprobs[token_id].logprob
                        yield (token_id, logprob_val)
                    else:
                        yield token_id
                    if stop_tokens is not None and token_id in stop_tokens:
                        break


@app.function(
    gpu=IMAGE_GPU,
    cpu=2.0,
    retries=0,
    image=img,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        VLLM_CACHE_DIR: vllm_cache_vol,
    },
    secrets=[modal.Secret.from_name("achatbot")],
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=1,
)
async def run(func, **kwargs):
    subprocess.run("nvidia-smi --version", shell=True)
    subprocess.run("nvcc --version", shell=True)
    subprocess.run("which vllm", shell=True)
    if torch.cuda.is_available():
        gpu_prop = torch.cuda.get_device_properties("cuda")
        print(gpu_prop)
        assert gpu_prop.major >= 8, f"now vllm gpt oss model just support cuda arch sm8.0+ :)"

    if asyncio.iscoroutinefunction(func):
        await func(**kwargs)
    else:
        func(**kwargs)


def get_tokenizer():
    import tiktoken

    o200k_base = tiktoken.get_encoding("o200k_base")
    tokenizer = tiktoken.Encoding(
        name="o200k_harmony",
        pat_str=o200k_base._pat_str,
        mergeable_ranks=o200k_base._mergeable_ranks,
        special_tokens={
            **o200k_base._special_tokens,
            "<|startoftext|>": 199998,
            "<|endoftext|>": 199999,
            "<|reserved_200000|>": 200000,
            "<|reserved_200001|>": 200001,
            "<|return|>": 200002,
            "<|constrain|>": 200003,
            "<|reserved_200004|>": 200004,
            "<|channel|>": 200005,
            "<|start|>": 200006,
            "<|end|>": 200007,
            "<|message|>": 200008,
            "<|reserved_200009|>": 200009,
            "<|reserved_200010|>": 200010,
            "<|reserved_200011|>": 200011,
            "<|call|>": 200012,
        }
        | {f"<|reserved_{i}|>": i for i in range(200013, 201088)},
    )
    return tokenizer


def tokenizer(**kwargs):
    tokenizer = get_tokenizer()
    stop_ids = tokenizer.encode("<|endoftext|>", allowed_special="all")
    print(stop_ids)
    assert tokenizer.eot_token == stop_ids[0]

    prompt = kwargs.get("prompt", "什么是快乐星球?")
    token_ids = tokenizer.encode(prompt)
    print(f"promt_token_ids: {token_ids}")
    tokens = tokenizer.decode(token_ids)
    print(f"promt_tokens: {tokens}")


def harmony_chat_tokenizer(**kwargs):
    tokenizer = get_tokenizer()

    # --- 1) Render the prefill with Harmony ---
    encoding = load_harmony_encoding(HarmonyEncodingName.HARMONY_GPT_OSS)

    reasoning_effort = kwargs.get("reasoning", "medium")
    model_identity = kwargs.get(
        "model_identity", "You are ChatGPT, a large language model trained by OpenAI."
    )
    convo = Conversation.from_messages(
        [
            Message.from_role_and_content(
                Role.SYSTEM,
                SystemContent.new()
                .with_reasoning_effort(REASONING_EFFORT[reasoning_effort])
                .with_model_identity(model_identity)
                .with_conversation_start_date(datetime.datetime.now().strftime("%Y-%m-%d")),
            ),
            Message.from_role_and_content(
                Role.DEVELOPER,
                DeveloperContent.new().with_instructions("Always respond in riddles"),
            ),
            Message.from_role_and_content(Role.USER, "What is the weather like in SF?"),
        ]
    )

    prompt_token_ids = encoding.render_conversation_for_completion(convo, Role.ASSISTANT)

    # Harmony stop tokens (pass to sampler so they won't be included in output)
    stop_token_ids = encoding.stop_tokens_for_assistant_actions()

    prompt_token = tokenizer.decode(prompt_token_ids)
    print(prompt_token_ids, prompt_token)

    print("\n" + "----" * 20 + "\n")

    stop_tokens = tokenizer.decode(stop_token_ids)
    print(stop_tokens, stop_token_ids)


def generate(**kwargs):
    # --- 1) Render the prefill with Harmony ---
    encoding = load_harmony_encoding(HarmonyEncodingName.HARMONY_GPT_OSS)

    convo = Conversation.from_messages(
        [
            Message.from_role_and_content(Role.SYSTEM, SystemContent.new()),
            Message.from_role_and_content(
                Role.DEVELOPER,
                DeveloperContent.new().with_instructions("Always respond in riddles"),
            ),
            Message.from_role_and_content(Role.USER, "What is the weather like in SF?"),
        ]
    )

    prefill_ids = encoding.render_conversation_for_completion(convo, Role.ASSISTANT)

    # Harmony stop tokens (pass to sampler so they won't be included in output)
    stop_token_ids = encoding.stop_tokens_for_assistant_actions()
    print(stop_token_ids)

    # --- 2) Run vLLM with prefill ---
    llm = LLM(
        model=model_path,
        tensor_parallel_size=int(os.getenv("TP", "1")),
        trust_remote_code=True,
        gpu_memory_utilization=0.7,
        quantization="mxfp4",
    )

    sampling = SamplingParams(
        temperature=0.6,
        top_p=0.95,
        max_tokens=128,
        repetition_penalty=1.1,
        stop_token_ids=stop_token_ids,
    )
    for i in range(2):
        start = perf_counter()
        outputs = llm.generate(
            prompt_token_ids=[prefill_ids],  # batch of size 1
            sampling_params=sampling,
        )
        print(f"{i} generate cost:{(perf_counter() - start):.3f} s")
        print(outputs)

        # vLLM gives you both text and token IDs
        gen = outputs[0].outputs[0]
        text = gen.text
        print(f"generate text: {text}")
        output_tokens = gen.token_ids  # <-- these are the completion token IDs (no prefill)

        # --- 3) Parse the completion token IDs back into structured Harmony messages ---
        entries = encoding.parse_messages_from_completion_tokens(output_tokens, Role.ASSISTANT)

        # 'entries' is a sequence of structured conversation entries (assistant messages, tool calls, etc.).
        for message in entries:
            print(json.dumps(message.to_dict(), indent=2))


def generate_stream(**kwargs):
    """
    no chat instruct, generate token step by step with tiktoken(o200k_harmony) encode/decode
    """
    generator = TokenGenerator(
        model_path=model_path, tensor_parallel_size=int(os.getenv("TP", "1"))
    )
    tokenizer = get_tokenizer()

    temperature = kwargs.get("temperature", 1.0)
    max_tokens = kwargs.get("max_tokens", 128)
    top_p = kwargs.get("top_p", 1.0)

    for i in range(3):  # compile, need warmup
        prompt = kwargs.get("prompt", "什么是快乐星球?")
        token_ids = tokenizer.encode(prompt)
        print(f"promt_token_ids: {token_ids}")

        times = []
        start = perf_counter()
        text = ""
        for token_id, logprob in generator.generate(
            token_ids,
            stop_tokens=[tokenizer.eot_token],
            temperature=temperature,
            top_p=top_p,
            max_tokens=max_tokens,
            return_logprobs=True,
        ):
            token_ids.append(token_id)
            decoded_token = tokenizer.decode(
                [token_id]
            )  # NOTE: unicode decode need stream parser, need to stop
            times.append(perf_counter() - start)
            text += decoded_token
            print(f"Generated token: {repr(decoded_token)}, logprob: {logprob}")
            start = perf_counter()

        print(f"{i} generated Text: {text}")
        print(f"{i} TTFT: {times[0]:.4f}s total time: {sum(times):.4f}s")


def chat_stream(**kwargs):
    """
    use harmony chat format to generate token step by step
    """

    # --- 1) Render the prefill with Harmony ---
    encoding = load_harmony_encoding(HarmonyEncodingName.HARMONY_GPT_OSS)

    reasoning_effort = kwargs.get("reasoning", "medium")
    model_identity = kwargs.get(
        "model_identity", "You are ChatGPT, a large language model trained by OpenAI."
    )
    convo = Conversation.from_messages(
        [
            Message.from_role_and_content(
                Role.SYSTEM,
                SystemContent.new()
                .with_reasoning_effort(REASONING_EFFORT[reasoning_effort])
                .with_model_identity(model_identity)
                .with_conversation_start_date(datetime.datetime.now().strftime("%Y-%m-%d")),
            ),
            Message.from_role_and_content(
                Role.DEVELOPER,
                DeveloperContent.new().with_instructions("Always respond in riddles"),
            ),
            Message.from_role_and_content(Role.USER, "What is the weather like in SF?"),
        ]
    )

    prompt_token_ids = encoding.render_conversation_for_completion(convo, Role.ASSISTANT)

    # Harmony stop tokens (pass to sampler so they won't be included in output)
    stop_token_ids = encoding.stop_tokens_for_assistant_actions()

    parser = StreamableParser(encoding, role=Role.ASSISTANT)
    generator = TokenGenerator(
        model_path=model_path, tensor_parallel_size=int(os.getenv("TP", "1"))
    )
    tokenizer = get_tokenizer()
    prompt_token = tokenizer.decode(prompt_token_ids)
    print(f"{prompt_token=}")

    field_created = False
    current_output_text = ""
    output_text_delta_buffer = ""
    for predicted_token in generator.generate(
        prompt_token_ids,
        stop_tokens=stop_token_ids,
        temperature=kwargs.get("temperature", 1.0),
        top_p=kwargs.get("top_p", 1.0),
        max_tokens=kwargs.get("max_tokens", 128),
    ):
        parser.process(predicted_token)

        if parser.state == StreamState.EXPECT_START:
            print("")  # new line
            field_created = False

        if not parser.last_content_delta:
            continue

        if not field_created:
            field_created = True
            if parser.current_channel == "final":
                print(termcolor.colored("Assistant:", "green"), flush=True)
            elif parser.current_recipient is not None:
                print(
                    termcolor.colored(f"Tool call to {parser.current_recipient}:", "cyan"),
                    flush=True,
                )
            else:
                print(termcolor.colored("CoT:", "yellow"), flush=True)

        should_send_output_text_delta = True
        output_text_delta_buffer += parser.last_content_delta
        if should_send_output_text_delta:
            print(output_text_delta_buffer, end="", flush=True)
            current_output_text += output_text_delta_buffer
            output_text_delta_buffer = ""
    print(f"{current_output_text=}")

    print(f"{parser.messages=}")


async def get_user_input(q: modal.Queue):
    rank = torch.distributed.get_rank() if torch.distributed.is_initialized() else 0
    if rank == 0:
        # user_input = input()
        user_input = await q.get.aio(partition="text")
    else:
        user_input = ""
    user_input_list = [user_input]
    if torch.distributed.is_initialized():
        torch.distributed.broadcast_object_list(user_input_list, 0)
    return user_input_list[0]


async def chat_tool_stream(**kwargs):
    generator = TokenGenerator(
        model_path=model_path, tensor_parallel_size=int(os.getenv("TP", "1"))
    )

    encoding = load_harmony_encoding(HarmonyEncodingName.HARMONY_GPT_OSS)

    reasoning_effort = kwargs.get("reasoning", "medium")
    model_identity = kwargs.get(
        "model_identity", "You are ChatGPT, a large language model trained by OpenAI."
    )
    system_message_content = (
        SystemContent.new()
        .with_reasoning_effort(REASONING_EFFORT[reasoning_effort])
        .with_model_identity(model_identity)
        .with_conversation_start_date(datetime.datetime.now().strftime("%Y-%m-%d"))
    )

    build_in_tool = kwargs.get("build_in_tool", "browser")

    if build_in_tool == "browser":
        backend = ExaBackend(
            source="web",
        )
        browser_tool = SimpleBrowserTool(backend=backend)
        system_message_content = system_message_content.with_tools(browser_tool.tool_config)

    if build_in_tool == "python":
        python_tool = PythonTool()
        system_message_content = system_message_content.with_tools(python_tool.tool_config)

    system_message = Message.from_role_and_content(Role.SYSTEM, system_message_content)
    messages = [system_message]

    is_apply_patch = kwargs.get("is_apply_patch", None)
    developer_message = kwargs.get("developer_message", "")
    if is_apply_patch:
        apply_patch_instructions = Path(apply_patch.__file__).parent / "apply_patch.md"
        developer_message = developer_message + "\n" if developer_message else ""
        developer_message += apply_patch_instructions.read_text()
        developer_message_content = (
            DeveloperContent.new()
            .with_instructions(developer_message)
            .with_function_tools(
                [
                    ToolDescription.new(
                        "apply_patch",
                        "Patch a file",
                        parameters={
                            "type": "string",
                            "description": "Formatted patch code",
                            "default": "*** Begin Patch\n*** End Patch\n",
                        },
                    ),
                ]
            )
        )
        messages.append(Message.from_role_and_content(Role.DEVELOPER, developer_message_content))
    else:
        developer_message_content = DeveloperContent.new().with_instructions(developer_message)
        messages.append(Message.from_role_and_content(Role.DEVELOPER, developer_message_content))

    raw = kwargs.get("raw", None)
    if raw:
        conversation = Conversation.from_messages(messages)
        tokens = encoding.render_conversation(conversation)
        system_message = encoding.decode(tokens)
        print(system_message, flush=True, end="")
        empty_user_message_tokens = encoding.render(Message.from_role_and_content(Role.USER, ""))
        user_message_start = encoding.decode(empty_user_message_tokens[:-1])
        user_message_end = encoding.decode(empty_user_message_tokens[-1:])
    else:
        # System message
        print(termcolor.colored("System Message:", "cyan"), flush=True)
        print(
            termcolor.colored("Model Identity:", "cyan"),
            system_message_content.model_identity,
            flush=True,
        )
        print(
            termcolor.colored("Reasoning Effort:", "cyan"),
            system_message_content.reasoning_effort,
            flush=True,
        )
        print(
            termcolor.colored("Conversation Start Date:", "cyan"),
            system_message_content.conversation_start_date,
            flush=True,
        )
        print(
            termcolor.colored("Knowledge Cutoff:", "cyan"),
            system_message_content.knowledge_cutoff,
            flush=True,
        )
        print(
            termcolor.colored("Browser Tool:", "cyan"),
            "Enabled" if build_in_tool == "browser" else "Disabled",
            flush=True,
        )
        print(
            termcolor.colored("Python Tool:", "cyan"),
            "Enabled" if build_in_tool == "python" else "Disabled",
            flush=True,
        )
        print(
            termcolor.colored("Apply Patch Function:", "cyan"),
            "Enabled" if is_apply_patch else "Disabled",
            flush=True,
        )
        # Developer message
        print(termcolor.colored("Developer Message:", "yellow"), flush=True)
        print(developer_message_content.instructions, flush=True)

    # Print the system message and the user message start
    MESSAGE_PADDING = 12
    q = kwargs.get("q", None)
    while True:
        last_message = messages[-1]
        if last_message.recipient is None:
            if raw:
                print(user_message_start, end="", flush=True)
                user_message = await get_user_input(q)
                print(user_message_end, flush=True, end="")
            else:
                print(termcolor.colored("User:".ljust(MESSAGE_PADDING), "red"), flush=True)
                user_message = await get_user_input(q)
            user_message = Message.from_role_and_content(Role.USER, user_message)
            messages.append(user_message)
        else:
            # Tool or function call
            if last_message.recipient.startswith("browser."):
                assert build_in_tool == "browser", "Browser tool is not enabled"
                tool_name = "Search"

                async def run_tool():
                    results = []
                    async for msg in browser_tool.process(last_message):
                        results.append(msg)
                    return results

                result = await run_tool()
                messages += result
            elif last_message.recipient.startswith("python"):
                assert build_in_tool == "python", "Python tool is not enabled"
                tool_name = "Python"

                async def run_tool():
                    results = []
                    async for msg in python_tool.process(last_message):
                        results.append(msg)
                    return results

                result = await run_tool()
                messages += result
            elif last_message.recipient == "functions.apply_patch":
                assert is_apply_patch, "Apply patch tool is not enabled"
                tool_name = "Apply Patch"
                text = last_message.content[0].text
                tool_output = None

                if text.startswith("{"):
                    # this is json, try to extract the patch from it
                    try:
                        some_dict = json.loads(text)
                        _, text = some_dict.popitem()
                    except Exception as e:
                        tool_output = f"Error parsing JSON: {e}"

                if tool_output is None:
                    try:
                        tool_output = apply_patch.apply_patch(text)
                    except Exception as e:
                        tool_output = f"Error applying patch: {e}"

                message = Message(
                    author=Author.new(Role.TOOL, last_message.recipient),
                    content=[TextContent(text=tool_output)],
                ).with_recipient("assistant")
                if last_message.channel:
                    message = message.with_channel(last_message.channel)

                result = [message]
                messages += result
            else:
                msg = f"Unknown tool or function call: {last_message=}"
                print(msg)
                # raise ValueError(f"Unknown tool or function call: {last_message.recipient}")

            # Print the tool or function call result
            if raw:
                rendered_result = encoding.render_conversation(Conversation.from_messages(result))
                print(encoding.decode(rendered_result), flush=True, end="")
            else:
                print(
                    termcolor.colored(f"{tool_name} output:".ljust(MESSAGE_PADDING), "magenta"),
                    flush=True,
                )
                show_browser_results = kwargs.get("show_browser_results", None)
                if tool_name == "Search" and not show_browser_results:
                    print("[Search results fed to the model]")
                else:
                    print(result[0].content[0].text)

        conversation = Conversation.from_messages(messages)
        tokens = encoding.render_conversation_for_completion(conversation, Role.ASSISTANT)

        if raw:
            # Print the last two tokens, which are the start of the assistant message
            print(encoding.decode(tokens[-2:]), flush=True, end="")

        parser = StreamableParser(encoding, role=Role.ASSISTANT)
        field_created = False
        current_output_text = ""
        output_text_delta_buffer = ""
        for predicted_token in generator.generate(
            tokens,
            stop_tokens=encoding.stop_tokens_for_assistant_actions(),
            temperature=kwargs.get("temperature", 1.0),
            top_p=kwargs.get("top_p", 1.0),
            max_tokens=kwargs.get("max_tokens", 2048),
        ):
            parser.process(predicted_token)
            if raw:
                print(encoding.decode([predicted_token]), end="", flush=True)
                continue

            if parser.state == StreamState.EXPECT_START:
                print("")  # new line
                field_created = False

            if not parser.last_content_delta:
                continue

            if not field_created:
                field_created = True
                if parser.current_channel == "final":
                    print(termcolor.colored("Assistant:", "green"), flush=True)
                elif parser.current_recipient is not None:
                    print(
                        termcolor.colored(f"Tool call to {parser.current_recipient}:", "cyan"),
                        flush=True,
                    )
                else:
                    print(termcolor.colored("CoT:", "yellow"), flush=True)

            should_send_output_text_delta = True
            output_text_delta_buffer += parser.last_content_delta
            if build_in_tool == "browser":
                updated_output_text, _annotations, has_partial_citations = (
                    browser_tool.normalize_citations(current_output_text + output_text_delta_buffer)
                )
                output_text_delta_buffer = updated_output_text[len(current_output_text) :]
                if has_partial_citations:
                    should_send_output_text_delta = False
            if should_send_output_text_delta:
                print(output_text_delta_buffer, end="", flush=True)
                current_output_text += output_text_delta_buffer
                output_text_delta_buffer = ""

        messages += parser.messages


@app.function(
    gpu=IMAGE_GPU,
    image=img,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        VLLM_CACHE_DIR: vllm_cache_vol,
    },
    # secrets=[modal.Secret.from_name("achatbot")],
    timeout=86400,  # default 300s
    # max_containers=1,
)
@modal.web_server(port=8000, startup_timeout=60 * 60)
@modal.concurrent(max_inputs=100, target_inputs=4)
def serve():
    """
    modal + vllm :
    - https://modal.com/docs/examples/vllm_inference
    - https://modal.com/llm-almanac/advisor
    """
    cmd = f"""
    VLLM_ATTENTION_BACKEND=TRITON_ATTN_VLLM_V1 vllm serve {model_path} --async-scheduling --served-model-name {model_name} --trust_remote_code --port 8000 --tensor-parallel-size {os.getenv("TP", "1")}
    """
    subprocess.Popen(cmd, shell=True, env=os.environ)


"""
# download hf tranformers weight(safetensors) for vllm to load
modal run src/download_models.py --repo-ids "openai/gpt-oss-20b"
modal run src/download_models.py --repo-ids "openai/gpt-oss-120b" --ignore-patterns "*.pt|*.bin|*original*|*metal*"

# see help
modal run src/llm/vllm/openai_gpt_oss.py --help 


# tokenizer
modal run src/llm/vllm/openai_gpt_oss.py --task tokenizer
modal run src/llm/vllm/openai_gpt_oss.py --task harmony_chat_tokenizer

# generate
IMAGE_GPU=A100 modal run src/llm/vllm/openai_gpt_oss.py --task generate
IMAGE_GPU=L40s modal run src/llm/vllm/openai_gpt_oss.py --task generate
IMAGE_GPU=H100 modal run src/llm/vllm/openai_gpt_oss.py --task generate

IMAGE_GPU=A100 modal run src/llm/vllm/openai_gpt_oss.py --task generate_stream
IMAGE_GPU=L40s modal run src/llm/vllm/openai_gpt_oss.py --task generate_stream
IMAGE_GPU=H100 modal run src/llm/vllm/openai_gpt_oss.py --task generate_stream

# chat
IMAGE_GPU=A100 modal run src/llm/vllm/openai_gpt_oss.py --task chat_stream
IMAGE_GPU=L40s modal run src/llm/vllm/openai_gpt_oss.py --task chat_stream
IMAGE_GPU=H100 modal run src/llm/vllm/openai_gpt_oss.py --task chat_stream

# local input --- queue --> remote to loop chat

## use browser tool(find,open,search), need env EXA_API_KEY from https://exa.ai
IMAGE_GPU=L40s modal run src/llm/vllm/openai_gpt_oss.py --task chat_tool_stream --build-in-tool browser
IMAGE_GPU=H100 modal run src/llm/vllm/openai_gpt_oss.py --task chat_tool_stream --build-in-tool browser
IMAGE_GPU=L40s modal run src/llm/vllm/openai_gpt_oss.py --task chat_tool_stream \
    --max-tokens 2048 --temperature=1.0 --top-p=1.0 \
    --build-in-tool browser --show-browser-results --model-identity "你是一名聊天助手，请用中文回复。"
IMAGE_GPU=H100 modal run src/llm/vllm/openai_gpt_oss.py --task chat_tool_stream \
    --max-tokens 2048 --temperature=1.0 --top-p=1.0 \
    --build-in-tool browser --show-browser-results --model-identity "你是一名聊天助手，请用中文回复。"
IMAGE_GPU=H100 modal run src/llm/vllm/openai_gpt_oss.py --task chat_tool_stream --build-in-tool browser --is-apply-patch --show-browser-results
IMAGE_GPU=H100 modal run src/llm/vllm/openai_gpt_oss.py --task chat_tool_stream --build-in-tool browser --raw --is-apply-patch --show-browser-results

## need python tool to run script need change python docker, u can change python tools to do local env python or use serverless function 
IMAGE_GPU=H100 modal run src/llm/vllm/openai_gpt_oss.py --task chat_tool_stream --build-in-tool python
IMAGE_GPU=H100 modal run src/llm/vllm/openai_gpt_oss.py --task chat_tool_stream --build-in-tool python --is-apply-patch 
IMAGE_GPU=H100 modal run src/llm/vllm/openai_gpt_oss.py --task chat_tool_stream --build-in-tool python --raw --is-apply-patch

# run vllm serve
IMAGE_GPU=H100 modal serve src/llm/vllm/openai_gpt_oss.py 
IMAGE_GPU=H200 modal serve src/llm/vllm/openai_gpt_oss.py 
# first ping to start serve
curl -XGET "https://weedge--openai-gpt-oss-serve-dev.modal.run/ping"
```
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /openapi.json, Methods: HEAD, GET
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /docs, Methods: HEAD, GET
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /docs/oauth2-redirect, Methods: HEAD, GET
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /redoc, Methods: HEAD, GET
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /health, Methods: GET
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /load, Methods: GET
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /ping, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /ping, Methods: GET
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /tokenize, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /detokenize, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /v1/models, Methods: GET
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /version, Methods: GET
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /v1/responses, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /v1/responses/{response_id}, Methods: GET
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /v1/responses/{response_id}/cancel, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /v1/chat/completions, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /v1/completions, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /v1/embeddings, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /pooling, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /classify, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /score, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /v1/score, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /v1/audio/transcriptions, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /v1/audio/translations, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /rerank, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /v1/rerank, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /v2/rerank, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /scale_elastic_ep, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /is_scaling_elastic_ep, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /invocations, Methods: POST
(APIServer pid=28) INFO 08-09 07:26:21 [launcher.py:37] Route: /metrics, Methods: GET
```
"""


def run_async_in_thread(coroutine, **kwargs):
    def run_coro_in_thread():
        asyncio.run(coroutine(**kwargs))

    thread = threading.Thread(target=run_coro_in_thread)
    thread.start()
    return thread


async def loop_input(q: modal.Queue):
    while True:
        text = input()
        print(f"local input: {text}")
        await q.put.aio(text, partition="text")


@app.local_entrypoint()
def main(
    task: str = "tokenizer",
    reasoning: str = "medium",  # low medium high
    model_identity: str = "You are ChatGPT, a large language model trained by OpenAI.",
    prompt: str = "什么是快乐星球?",
    temperature: float = 1.0,
    top_p: float = 1.0,
    max_tokens: int = 128,
    # for chat_tool_stream
    build_in_tool: str = "browser",  # build-in tools: browser,python
    is_apply_patch: bool = False,  # Make apply_patch tool available to the model (default: False)
    show_browser_results: bool = False,  # Show browser results (default: False)
    developer_message: str = "",  # Developer message (default: )
    raw: bool = False,  # Raw mode (does not render Harmony encoding) (default: False)
):
    print(task)
    tasks = {
        "tokenizer": tokenizer,
        "harmony_chat_tokenizer": harmony_chat_tokenizer,
        "generate": generate,
        "generate_stream": generate_stream,
        "chat_stream": chat_stream,
        "chat_tool_stream": chat_tool_stream,
    }
    if task not in tasks:
        raise ValueError(f"task {task} not found")
    print(f"running task {task}")
    with modal.Queue.ephemeral() as q:
        if task == "chat_tool_stream":
            run_async_in_thread(loop_input, q=q)

        run.remote(
            tasks[task],
            reasoning=reasoning.lower(),
            model_identity=model_identity,
            prompt=prompt,
            temperature=temperature,
            top_p=top_p,
            max_tokens=max_tokens,
            build_in_tool=build_in_tool,
            is_apply_patch=is_apply_patch,
            show_browser_results=show_browser_results,
            developer_message=developer_message,
            raw=raw,
            q=q,
        )



import pyaudio
import grpc
import json
import webrtcvad
import time
import asyncio
import argparse

from grpc_client import transcribe_audio_bytes
from paraformer_pb2_grpc import ASRStub

async def deal_chunk(sig_mic):
    global stub,SPEAKING,asr_user,language,sample_rate
    if vad.is_speech(sig_mic, sample_rate): #speaking
        SPEAKING = True
        response = transcribe_audio_bytes(stub, sig_mic, user=asr_user, language=language, speaking = True, isEnd = False) #speaking, send audio to server.
    else: #silence   
        begin_time = 0
        if SPEAKING: #means we have some audio recorded, send recognize order to server.
            SPEAKING = False
            begin_time = int(round(time.time() * 1000))            
            response = transcribe_audio_bytes(stub, None, user=asr_user, language=language, speaking = False, isEnd = False) #speak end, call server for recognize one sentence
            resp = response.next()           
            if "decoding" == resp.action:   
                resp = response.next() #TODO, blocking operation may leads to miss some audio clips. C++ multi-threading is preferred.
                if "finish" == resp.action:        
                    end_time = int(round(time.time() * 1000))
                    print (json.loads(resp.sentence))
                    print ("delay in ms: %d " % (end_time - begin_time))
                else:
                    pass
        

async def record(host,port,sample_rate,mic_chunk,record_seconds,asr_user,language):
    with grpc.insecure_channel('{}:{}'.format(host, port)) as channel:
        global stub
        stub = ASRStub(channel)
        for i in range(0, int(sample_rate / mic_chunk * record_seconds)):
     
            sig_mic = stream.read(mic_chunk,exception_on_overflow = False) 
            await asyncio.create_task(deal_chunk(sig_mic))

        #end grpc
        response = transcribe_audio_bytes(stub, None, user=asr_user, language=language, speaking = False, isEnd = True)
        print (response.next().action)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--host",
                        type=str,
                        default="127.0.0.1",
                        required=True,
                        help="grpc server host ip")
                        
    parser.add_argument("--port",
                        type=int,
                        default=10095,
                        required=True,
                        help="grpc server port")              
                        
    parser.add_argument("--user_allowed",
                        type=str,
                        default="project1_user1",
                        help="allowed user for grpc client")
                        
    parser.add_argument("--sample_rate",
                        type=int,
                        default=16000,
                        help="audio sample_rate from client")    

    parser.add_argument("--mic_chunk",
                        type=int,
                        default=160,
                        help="chunk size for mic")  

    parser.add_argument("--record_seconds",
                        type=int,
                        default=120,
                        help="run specified seconds then exit ")                       

    args = parser.parse_args()
    

    SPEAKING = False
    asr_user = args.user_allowed
    sample_rate = args.sample_rate
    language = 'zh-CN'  
    

    vad = webrtcvad.Vad()
    vad.set_mode(1)

    FORMAT = pyaudio.paInt16
    CHANNELS = 1
    p = pyaudio.PyAudio()
    
    stream = p.open(format=FORMAT,
                channels=CHANNELS,
                rate=args.sample_rate,
                input=True,
                frames_per_buffer=args.mic_chunk)
                
    print("* recording")
    asyncio.run(record(args.host,args.port,args.sample_rate,args.mic_chunk,args.record_seconds,args.user_allowed,language))
    stream.stop_stream()
    stream.close()
    p.terminate()
    print("recording stop")

    




import os
import ctypes
import pathlib


whisper_cpp_dir = os.getenv("WHISPER_CPP_DIR", "/Users/wuyong/project/pywhispercpp/whisper.cpp")
library_path = os.path.join(whisper_cpp_dir, "build/src/libwhisper.dylib")
whisper_model_path = "./models/ggml-base.bin"
vad_model_path = "./models/ggml-silero-v5.1.2.bin"
wav_file = os.path.join(whisper_cpp_dir, "samples/jfk.wav")


# this needs to match the C struct in whisper.h
class WhisperFullParams(ctypes.Structure):
    _fields_ = [
        ("strategy", ctypes.c_int),
        #
        ("n_max_text_ctx", ctypes.c_int),
        ("n_threads", ctypes.c_int),
        ("offset_ms", ctypes.c_int),
        ("duration_ms", ctypes.c_int),
        #
        ("translate", ctypes.c_bool),
        ("no_context", ctypes.c_bool),
        ("single_segment", ctypes.c_bool),
        ("print_special", ctypes.c_bool),
        ("print_progress", ctypes.c_bool),
        ("print_realtime", ctypes.c_bool),
        ("print_timestamps", ctypes.c_bool),
        #
        ("token_timestamps", ctypes.c_bool),
        ("thold_pt", ctypes.c_float),
        ("thold_ptsum", ctypes.c_float),
        ("max_len", ctypes.c_int),
        ("max_tokens", ctypes.c_int),
        #
        ("speed_up", ctypes.c_bool),
        ("audio_ctx", ctypes.c_int),
        #
        ("prompt_tokens", ctypes.c_void_p),
        ("prompt_n_tokens", ctypes.c_int),
        #
        ("language", ctypes.c_char_p),
        #
        ("suppress_blank", ctypes.c_bool),
        #
        ("temperature_inc", ctypes.c_float),
        ("entropy_thold", ctypes.c_float),
        ("logprob_thold", ctypes.c_float),
        ("no_speech_thold", ctypes.c_float),
        #
        ("greedy", ctypes.c_int * 1),
        ("beam_search", ctypes.c_int * 3),
        #
        ("new_segment_callback", ctypes.c_void_p),
        ("new_segment_callback_user_data", ctypes.c_void_p),
        #
        ("encoder_begin_callback", ctypes.c_void_p),
        ("encoder_begin_callback_user_data", ctypes.c_void_p),
    ]


def demo():
    # load library and model
    whisper = ctypes.CDLL(library_path)

    # tell Python what are the return types of the functions
    whisper.whisper_init_from_file.restype = ctypes.c_void_p
    whisper.whisper_full_default_params.restype = WhisperFullParams
    whisper.whisper_full_get_segment_text.restype = ctypes.c_char_p

    # initialize whisper.cpp context
    ctx = whisper.whisper_init_from_file(whisper_model_path.encode("utf-8"))

    # get default whisper parameters and adjust as needed
    params = whisper.whisper_full_default_params()
    print(params)
    params.print_realtime = True
    params.print_progress = False

    # load WAV file
    # this is needed to read the WAV file properly
    from scipy.io import wavfile

    samplerate, data = wavfile.read(wav_file)

    # convert to 32-bit float
    data = data.astype("float32") / 32768.0

    # run the inference
    result = whisper.whisper_full(
        ctypes.c_void_p(ctx),
        params,
        data.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
        len(data),
    )
    if result != 0:
        print("Error: {}".format(result))
        exit(1)

    # print results from Python
    # print("\nResults from Python:\n")
    n_segments = whisper.whisper_full_n_segments(ctypes.c_void_p(ctx))
    for i in range(n_segments):
        t0 = whisper.whisper_full_get_segment_t0(ctypes.c_void_p(ctx), i)
        t1 = whisper.whisper_full_get_segment_t1(ctypes.c_void_p(ctx), i)
        txt = whisper.whisper_full_get_segment_text(ctypes.c_void_p(ctx), i)

        print(f"{t0 / 1000.0:.3f} - {t1 / 1000.0:.3f} : {txt.decode('utf-8')}")

    # free the memory
    whisper.whisper_free(ctypes.c_void_p(ctx))


def transcribe():
    import soundfile as sf

    from whispercpy import WhisperCPP
    from whispercpy.utils import to_timestamp

    # reading audio
    data, sr = sf.read(wav_file, dtype="float32")

    # load model
    model = WhisperCPP(
        library_path, whisper_model_path, vad_model_path, use_gpu=False, verbose=True
    )

    print("--------- VAD Result ----------")

    # get vad results
    for segment in model.vad(data):
        print(
            f"[{to_timestamp(segment.t0, False)}" + " --> " + f"{to_timestamp(segment.t1, False)}]"
        )

    print("--------- ASR Result ----------")

    # get asr results
    for segment in model.transcribe(data, language="en", beam_size=5, token_timestamps=True):
        print(f"{segment=}")
        print(
            f"[{to_timestamp(segment.t0, False)}"
            + " --> "
            + f"{to_timestamp(segment.t1, False)}] "
            + segment.text
        )
        print("--------- Token Info ----------")
        print(
            "\n".join(
                [
                    f"[{to_timestamp(token.t0, False)}"
                    + " --> "
                    + f"{to_timestamp(token.t1, False)}] {token.text}"
                    for token in segment.tokens
                ]
            )
        )
        print("-------------------------------")


count = 0


def live():
    import sounddevice as sd

    from whispercpy import WhisperCPP, WhisperStream
    from whispercpy.utils import to_timestamp
    from whispercpy.constant import STREAMING_ENDING

    core = WhisperCPP(library_path, whisper_model_path, use_gpu=False)
    asr = WhisperStream(core, language="zh", return_token=True)

    samplerate = 16000
    block_duration = 0.25
    block_size = int(samplerate * block_duration)
    channels = 1

    def callback(indata, frames, time, status):
        global count

        chunk = indata.copy().tobytes()
        asr.pipe(chunk)
        transcript = asr.get_transcript()
        transcripts = asr.get_transcripts()

        if len(transcripts) > count:
            print("\r" + transcripts[-1].text)
            print("--")
            count += 1
        else:
            print(f"\r{transcript.text}", end="", flush=True)

    def print_result():
        transcripts = asr.get_transcripts()

        for transcript in transcripts:
            print(
                f"[{to_timestamp(transcript.t0, False)}"
                + " --> "
                + f"{to_timestamp(transcript.t1, False)}] "
                + transcript.text
            )
            print("-------------------------------")
            print(
                "\n".join(
                    [
                        f"[{to_timestamp(token.t0, False)}"
                        + " --> "
                        + f"{to_timestamp(token.t1, False)}] {token.text}"
                        for token in transcript.tokens
                    ]
                )
            )
            print("-------------------------------")

    # Recording
    try:
        with sd.InputStream(
            samplerate=samplerate,
            channels=channels,
            callback=callback,
            blocksize=block_size,
            dtype="float32",
        ):
            print("🎤 Recording for ASR... Press Ctrl+C to stop.")
            while True:
                sd.sleep(1000)
    except KeyboardInterrupt:
        print("⏹️ Recording stopped.")
        # send end signal, and await
        asr.pipe(STREAMING_ENDING).join()
        print_result()


if __name__ == "__main__":
    # demo()
    transcribe()
    # live()



import zlib
from typing import Iterator, TextIO


def exact_div(x, y):
    assert x % y == 0
    return x // y


def str2bool(string):
    str2val = {"True": True, "False": False}
    if string in str2val:
        return str2val[string]
    else:
        raise ValueError(f"Expected one of {set(str2val.keys())}, got {string}")


def optional_int(string):
    return None if string == "None" else int(string)


def optional_float(string):
    return None if string == "None" else float(string)


def compression_ratio(text) -> float:
    return len(text) / len(zlib.compress(text.encode("utf-8")))


def format_timestamp(seconds: float, always_include_hours: bool = False, decimal_marker: str = '.'):
    assert seconds >= 0, "non-negative timestamp expected"
    milliseconds = round(seconds * 1000.0)

    hours = milliseconds // 3_600_000
    milliseconds -= hours * 3_600_000

    minutes = milliseconds // 60_000
    milliseconds -= minutes * 60_000

    seconds = milliseconds // 1_000
    milliseconds -= seconds * 1_000

    hours_marker = f"{hours:02d}:" if always_include_hours or hours > 0 else ""
    return f"{hours_marker}{minutes:02d}:{seconds:02d}{decimal_marker}{milliseconds:03d}"


def write_txt(transcript: Iterator[dict], file: TextIO):
    for segment in transcript:
        print(segment['text'].strip(), file=file, flush=True)


def write_vtt(transcript: Iterator[dict], file: TextIO):
    print("WEBVTT\n", file=file)
    for segment in transcript:
        print(
            f"{format_timestamp(segment['start'])} --> {format_timestamp(segment['end'])}\n"
            f"{segment['text'].strip().replace('-->', '->')}\n",
            file=file,
            flush=True,
        )


def write_srt(transcript: Iterator[dict], file: TextIO):
    """
    Write a transcript to a file in SRT format.

    Example usage:
        from pathlib import Path
        from whisper.utils import write_srt

        result = transcribe(model, audio_path, temperature=temperature, **args)

        # save SRT
        audio_basename = Path(audio_path).stem
        with open(Path(output_dir) / (audio_basename + ".srt"), "w", encoding="utf-8") as srt:
            write_srt(result["segments"], file=srt)
    """
    for i, segment in enumerate(transcript, start=1):
        # write srt lines
        print(
            f"{i}\n"
            f"{format_timestamp(segment['start'], always_include_hours=True, decimal_marker=',')} --> "
            f"{format_timestamp(segment['end'], always_include_hours=True, decimal_marker=',')}\n"
            f"{segment['text'].strip().replace('-->', '->')}\n",
            file=file,
            flush=True,
        )



import numpy as np
import torch
import torch.nn.functional as F
from itertools import permutations
from torch import nn


def standard_loss(ys, ts, label_delay=0):
    losses = [F.binary_cross_entropy(torch.sigmoid(y), t) * len(y) for y, t in zip(ys, ts)]
    loss = torch.sum(torch.stack(losses))
    n_frames = torch.from_numpy(np.array(np.sum([t.shape[0] for t in ts]))).to(torch.float32).to(ys[0].device)
    loss = loss / n_frames
    return loss


def batch_pit_n_speaker_loss(ys, ts, n_speakers_list):
    max_n_speakers = ts[0].shape[1]
    olens = [y.shape[0] for y in ys]
    ys = nn.utils.rnn.pad_sequence(ys, batch_first=True, padding_value=-1)
    ys_mask = [torch.ones(olen).to(ys.device) for olen in olens]
    ys_mask = torch.nn.utils.rnn.pad_sequence(ys_mask, batch_first=True, padding_value=0).unsqueeze(-1)

    losses = []
    for shift in range(max_n_speakers):
        ts_roll = [torch.roll(t, -shift, dims=1) for t in ts]
        ts_roll = nn.utils.rnn.pad_sequence(ts_roll, batch_first=True, padding_value=-1)
        loss = F.binary_cross_entropy(torch.sigmoid(ys), ts_roll, reduction='none')
        if ys_mask is not None:
            loss = loss * ys_mask
        loss = torch.sum(loss, dim=1)
        losses.append(loss)
    losses = torch.stack(losses, dim=2)

    perms = np.array(list(permutations(range(max_n_speakers)))).astype(np.float32)
    perms = torch.from_numpy(perms).to(losses.device)
    y_ind = torch.arange(max_n_speakers, dtype=torch.float32, device=losses.device)
    t_inds = torch.fmod(perms - y_ind, max_n_speakers).to(torch.long)

    losses_perm = []
    for t_ind in t_inds:
        losses_perm.append(
            torch.mean(losses[:, y_ind.to(torch.long), t_ind], dim=1))
    losses_perm = torch.stack(losses_perm, dim=1)

    def select_perm_indices(num, max_num):
        perms = list(permutations(range(max_num)))
        sub_perms = list(permutations(range(num)))
        return [
            [x[:num] for x in perms].index(perm)
            for perm in sub_perms]

    masks = torch.full_like(losses_perm, device=losses.device, fill_value=float('inf'))
    for i, t in enumerate(ts):
        n_speakers = n_speakers_list[i]
        indices = select_perm_indices(n_speakers, max_n_speakers)
        masks[i, indices] = 0
    losses_perm += masks

    min_loss = torch.sum(torch.min(losses_perm, dim=1)[0])
    n_frames = torch.from_numpy(np.array(np.sum([t.shape[0] for t in ts]))).to(losses.device)
    min_loss = min_loss / n_frames

    min_indices = torch.argmin(losses_perm, dim=1)
    labels_perm = [t[:, perms[idx].to(torch.long)] for t, idx in zip(ts, min_indices)]
    labels_perm = [t[:, :n_speakers] for t, n_speakers in zip(labels_perm, n_speakers_list)]

    return min_loss, labels_perm



import os
import json
import logging

import librosa
import torch

from .vocoder.bigvgan import BigVGAN
from .vocoder.utils import get_melspec, AttrDict, load_checkpoint

logger = logging.getLogger(__name__)


class BigVGANWrapper:
    def __init__(
        self, vocoder: BigVGAN, device: torch.device, h: AttrDict, dtype=None
    ) -> None:
        self.vocoder = vocoder.to(device)
        if dtype is not None:
            self.vocoder = self.vocoder.to(dtype)
        self.vocoder = self.vocoder.eval()
        self.device = device
        self.h = h

    def to_dtype(self, dtype):
        self.vocoder = self.vocoder.to(dtype)

    def extract_mel_from_wav(self, wav_path=None, wav_data=None):
        """
        params:
            wav_path: str, path of the wav, should be 24k
            wav_data: torch.tensor or numpy array, shape [T], wav data, should be 24k
        return:
            mel: [T, num_mels], torch.tensor
        """
        if wav_data is None:
            wav_data, _ = librosa.load(wav_path, sr=self.h["sampling_rate"])

        wav_data = torch.tensor(wav_data).unsqueeze(0)

        mel = get_melspec(
            y=wav_data,
            n_fft=self.h["n_fft"],
            num_mels=self.h["num_mels"],
            sampling_rate=self.h["sampling_rate"],
            hop_size=self.h["hop_size"],
            win_size=self.h["win_size"],
            fmin=self.h["fmin"],
            fmax=self.h["fmax"],
        )
        return mel.squeeze(0).transpose(0, 1)

    @torch.inference_mode()
    def extract_mel_from_wav_batch(self, wav_data):
        """
        params:
            wav_data: torch.tensor or numpy array, shape [Batch, T], wav data, should be 24k
        return:
            mel: [Batch, T, num_mels], torch.tensor
        """

        wav_data = torch.tensor(wav_data)

        mel = get_melspec(
            wav=wav_data,
            n_fft=self.h["n_fft"],
            num_mels=self.h["num_mels"],
            sampling_rate=self.h["sampling_rate"],
            hop_size=self.h["hop_size"],
            win_size=self.h["win_size"],
            fmin=self.h["fmin"],
            fmax=self.h["fmax"],
        )
        return mel.transpose(1, 2)

    def decode_mel(self, mel):
        """
        params:
            mel: [T, num_mels], torch.tensor
        return:
            wav: [1, T], torch.tensor
        """
        mel = mel.transpose(0, 1).unsqueeze(0).to(self.device)
        wav = self.vocoder(mel)
        return wav.squeeze(0)

    def decode_mel_batch(self, mel):
        """
        params:
            mel: [B, T, num_mels], torch.tensor
        return:
            wav: [B, 1, T], torch.tensor
        """
        mel = mel.transpose(1, 2).to(self.device)
        wav = self.vocoder(mel)
        return wav

    @classmethod
    def from_pretrained(cls, model_config, ckpt_path, device):
        with open(model_config) as f:
            data = f.read()
        json_config = json.loads(data)
        h = AttrDict(json_config)
        vocoder = BigVGAN(h, True)
        state_dict_g = load_checkpoint(ckpt_path, "cpu")
        vocoder.load_state_dict(state_dict_g["generator"])

        logger.info(">>> Load vocoder from {}".format(ckpt_path))
        return cls(vocoder, device, h)



from collections.abc import Generator, Iterable
import os

from youtube_transcript_api import YouTubeTranscriptApi
from pydantic import BaseModel, Field
import instructor
import google.generativeai as genai


# Load environment variables from .env file
from dotenv import load_dotenv

load_dotenv(override=True)

genai.configure(api_key=os.environ["GOOGLE_API_KEY"])
client = instructor.from_gemini(
    client=genai.GenerativeModel(
        model_name="models/gemini-1.5-flash-latest",
    ),
    mode=instructor.Mode.GEMINI_JSON,
)


def extract_video_id(url: str) -> str | None:
    import re

    match = re.search(r"v=([a-zA-Z0-9_-]+)", url)
    if match:
        return match.group(1)


class TranscriptSegment(BaseModel):
    source_id: int
    start: float
    text: str


def get_transcript_with_timing(
    video_id: str, languages=("en", "zh-CN")
) -> Generator[TranscriptSegment, None, None]:
    """
    Fetches the transcript of a YouTube video along with the start and end times for each text segment,
    and returns them as a list of Pydantic models.

    Parameters:
    - video_id (str): The YouTube video ID for which the transcript is to be fetched.

    Returns:
    - A generator that yields TranscriptSegment models, each containing 'index', 'start', and 'text' keys.
    """
    transcript = YouTubeTranscriptApi.get_transcript(video_id, languages=languages)
    for ii, segment in enumerate(transcript):
        yield TranscriptSegment(source_id=ii, start=segment["start"], text=segment["text"])


class YoutubeClip(BaseModel):
    title: str = Field(description="Specific and informative title for the individual clip.")
    description: str = Field(
        description="A detailed description of the clip, including any notable quotes or phrases. should be a summary of sorts."
    )
    start: float
    end: float
    source_ids: list[int] = Field(exclude=True)


class YoutubeClips(BaseModel):
    clips: list[YoutubeClip]


def yield_clips(segments: Iterable[TranscriptSegment]) -> Iterable[YoutubeClips]:
    """
    Extracts a list of YouTube clips from a list of transcript segments.

    Parameters:
    - segments (Iterable[TranscriptSegment]): A list of TranscriptSegment models, each containing 'index', 'start', and 'text' keys.

    Returns:
    - A generator that yields YoutubeClipw models, each containing 'title', 'description', 'start', 'end', and 'source_ids' keys.
    """

    return client.chat.completions.create(
        stream=True,
        messages=[
            {
                "role": "system",
                "content": "You are given a sequence of YouTube transcripts and your job is to return notable clips that can be recut as smaller videos. give very specific titles and descriptions. Make sure the length of clips is proportional to the length of the video. Note that this is a transcript and so there might be spelling errors. Note that and correct any spellings. Use the context to make sure you're spelling things correctly. ",
            },
            {
                "role": "user",
                "content": f"Let's use the following transcript segments.\n{segments}",
            },
        ],
        response_model=instructor.Partial[YoutubeClips],
        validation_context={
            "segments": segments,
        },
    )  # type: ignore


# Example usage
if __name__ == "__main__":
    from rich.table import Table
    from rich.console import Console
    from rich.prompt import Prompt

    console = Console()
    url = Prompt.ask("Enter a YouTube URL")

    with console.status("[bold green]Processing YouTube URL...") as status:
        video_id = extract_video_id(url)

        if video_id is None:
            raise ValueError("Invalid YouTube video URL")

        transcript = list(get_transcript_with_timing(video_id))
        status.update("[bold green]Generating clips...")

        for clip in yield_clips(transcript):
            console.clear()

            table = Table(title="YouTube Clips", padding=(0, 1))

            table.add_column("Title", style="cyan")
            table.add_column("Description", style="magenta")
            table.add_column("Start", justify="right", style="green")
            table.add_column("End", justify="right", style="green")
            for youtube_clip in clip.clips or []:
                table.add_row(
                    youtube_clip.title,
                    youtube_clip.description,
                    str(youtube_clip.start),
                    str(youtube_clip.end),
                )
            console.print(table)




import torch
import numpy as np
import math
from funasr_local.modules.nets_utils import make_pad_mask
import logging
import torch.nn.functional as F
from funasr_local.modules.streaming_utils.utils import sequence_mask



class overlap_chunk():
	"""
	Author: Speech Lab of DAMO Academy, Alibaba Group
	San-m: Memory equipped self-attention for end-to-end speech recognition
	https://arxiv.org/abs/2006.01713

	"""
	def __init__(self,
		chunk_size: tuple = (16,),
		stride: tuple = (10,),
		pad_left: tuple = (0,),
		encoder_att_look_back_factor: tuple = (1,),
        shfit_fsmn: int = 0,
        decoder_att_look_back_factor: tuple = (1,),
	):

		pad_left = self.check_chunk_size_args(chunk_size, pad_left)
		encoder_att_look_back_factor = self.check_chunk_size_args(chunk_size, encoder_att_look_back_factor)
		decoder_att_look_back_factor = self.check_chunk_size_args(chunk_size, decoder_att_look_back_factor)
		self.chunk_size, self.stride, self.pad_left, self.encoder_att_look_back_factor, self.decoder_att_look_back_factor \
			= chunk_size, stride, pad_left, encoder_att_look_back_factor, decoder_att_look_back_factor
		self.shfit_fsmn = shfit_fsmn
		self.x_add_mask = None
		self.x_rm_mask = None
		self.x_len = None
		self.mask_shfit_chunk = None
		self.mask_chunk_predictor = None
		self.mask_att_chunk_encoder = None
		self.mask_shift_att_chunk_decoder = None
		self.chunk_outs = None
		self.chunk_size_cur, self.stride_cur, self.pad_left_cur, self.encoder_att_look_back_factor_cur, self.chunk_size_pad_shift_cur \
			= None, None, None, None, None

	def check_chunk_size_args(self, chunk_size, x):
		if len(x) < len(chunk_size):
			x = [x[0] for i in chunk_size]
		return x

	def get_chunk_size(self,
		ind: int = 0
	):
		# with torch.no_grad:
		chunk_size, stride, pad_left, encoder_att_look_back_factor, decoder_att_look_back_factor = \
			self.chunk_size[ind], self.stride[ind], self.pad_left[ind], self.encoder_att_look_back_factor[ind], self.decoder_att_look_back_factor[ind]
		self.chunk_size_cur, self.stride_cur, self.pad_left_cur, self.encoder_att_look_back_factor_cur, self.chunk_size_pad_shift_cur, self.decoder_att_look_back_factor_cur \
			= chunk_size, stride, pad_left, encoder_att_look_back_factor, chunk_size + self.shfit_fsmn, decoder_att_look_back_factor
		return self.chunk_size_cur, self.stride_cur, self.pad_left_cur, self.encoder_att_look_back_factor_cur, self.chunk_size_pad_shift_cur

	def random_choice(self, training=True, decoding_ind=None):
		chunk_num = len(self.chunk_size)
		ind = 0
		if training and chunk_num > 1:
			ind = torch.randint(0, chunk_num-1, ()).cpu().item()
		if not training and decoding_ind is not None:
			ind = int(decoding_ind)

		return ind




	def gen_chunk_mask(self, x_len, ind=0, num_units=1, num_units_predictor=1):

		with torch.no_grad():
			x_len = x_len.cpu().numpy()
			x_len_max = x_len.max()

			chunk_size, stride, pad_left, encoder_att_look_back_factor, chunk_size_pad_shift = self.get_chunk_size(ind)
			shfit_fsmn = self.shfit_fsmn
			pad_right = chunk_size - stride - pad_left

			chunk_num_batch = np.ceil(x_len/stride).astype(np.int32)
			x_len_chunk = (chunk_num_batch-1) * chunk_size_pad_shift + shfit_fsmn + pad_left + 0 + x_len - (chunk_num_batch-1) * stride
			x_len_chunk = x_len_chunk.astype(x_len.dtype)
			x_len_chunk_max = x_len_chunk.max()

			chunk_num = int(math.ceil(x_len_max/stride))
			dtype = np.int32
			max_len_for_x_mask_tmp = max(chunk_size, x_len_max + pad_left)
			x_add_mask = np.zeros([0, max_len_for_x_mask_tmp], dtype=dtype)
			x_rm_mask = np.zeros([max_len_for_x_mask_tmp, 0], dtype=dtype)
			mask_shfit_chunk = np.zeros([0, num_units], dtype=dtype)
			mask_chunk_predictor = np.zeros([0, num_units_predictor], dtype=dtype)
			mask_shift_att_chunk_decoder = np.zeros([0, 1], dtype=dtype)
			mask_att_chunk_encoder = np.zeros([0, chunk_num*chunk_size_pad_shift], dtype=dtype)
			for chunk_ids in range(chunk_num):
				# x_mask add
				fsmn_padding = np.zeros((shfit_fsmn, max_len_for_x_mask_tmp), dtype=dtype)
				x_mask_cur = np.diag(np.ones(chunk_size, dtype=np.float32))
				x_mask_pad_left = np.zeros((chunk_size, chunk_ids * stride), dtype=dtype)
				x_mask_pad_right = np.zeros((chunk_size, max_len_for_x_mask_tmp), dtype=dtype)
				x_cur_pad = np.concatenate([x_mask_pad_left, x_mask_cur, x_mask_pad_right], axis=1)
				x_cur_pad = x_cur_pad[:chunk_size, :max_len_for_x_mask_tmp]
				x_add_mask_fsmn = np.concatenate([fsmn_padding, x_cur_pad], axis=0)
				x_add_mask = np.concatenate([x_add_mask, x_add_mask_fsmn], axis=0)

				# x_mask rm
				fsmn_padding = np.zeros((max_len_for_x_mask_tmp, shfit_fsmn),dtype=dtype)
				padding_mask_left = np.zeros((max_len_for_x_mask_tmp, pad_left),dtype=dtype)
				padding_mask_right = np.zeros((max_len_for_x_mask_tmp, pad_right), dtype=dtype)
				x_mask_cur = np.diag(np.ones(stride, dtype=dtype))
				x_mask_cur_pad_top = np.zeros((chunk_ids*stride, stride), dtype=dtype)
				x_mask_cur_pad_bottom = np.zeros((max_len_for_x_mask_tmp, stride), dtype=dtype)
				x_rm_mask_cur = np.concatenate([x_mask_cur_pad_top, x_mask_cur, x_mask_cur_pad_bottom], axis=0)
				x_rm_mask_cur = x_rm_mask_cur[:max_len_for_x_mask_tmp, :stride]
				x_rm_mask_cur_fsmn = np.concatenate([fsmn_padding, padding_mask_left, x_rm_mask_cur, padding_mask_right], axis=1)
				x_rm_mask = np.concatenate([x_rm_mask, x_rm_mask_cur_fsmn], axis=1)

				# fsmn_padding_mask
				pad_shfit_mask = np.zeros([shfit_fsmn, num_units], dtype=dtype)
				ones_1 = np.ones([chunk_size, num_units], dtype=dtype)
				mask_shfit_chunk_cur = np.concatenate([pad_shfit_mask, ones_1], axis=0)
				mask_shfit_chunk = np.concatenate([mask_shfit_chunk, mask_shfit_chunk_cur], axis=0)

				# predictor mask
				zeros_1 = np.zeros([shfit_fsmn + pad_left, num_units_predictor], dtype=dtype)
				ones_2 = np.ones([stride, num_units_predictor], dtype=dtype)
				zeros_3 = np.zeros([chunk_size - stride - pad_left, num_units_predictor], dtype=dtype)
				ones_zeros = np.concatenate([ones_2, zeros_3], axis=0)
				mask_chunk_predictor_cur = np.concatenate([zeros_1, ones_zeros], axis=0)
				mask_chunk_predictor = np.concatenate([mask_chunk_predictor, mask_chunk_predictor_cur], axis=0)

				# encoder att mask
				zeros_1_top = np.zeros([shfit_fsmn, chunk_num*chunk_size_pad_shift], dtype=dtype)

				zeros_2_num = max(chunk_ids - encoder_att_look_back_factor, 0)
				zeros_2 = np.zeros([chunk_size, zeros_2_num*chunk_size_pad_shift], dtype=dtype)

				encoder_att_look_back_num = max(chunk_ids - zeros_2_num, 0)
				zeros_2_left = np.zeros([chunk_size, shfit_fsmn], dtype=dtype)
				ones_2_mid = np.ones([stride, stride], dtype=dtype)
				zeros_2_bottom = np.zeros([chunk_size-stride, stride], dtype=dtype)
				zeros_2_right = np.zeros([chunk_size, chunk_size-stride], dtype=dtype)
				ones_2 = np.concatenate([ones_2_mid, zeros_2_bottom], axis=0)
				ones_2 = np.concatenate([zeros_2_left, ones_2, zeros_2_right], axis=1)
				ones_2 = np.tile(ones_2, [1, encoder_att_look_back_num])

				zeros_3_left = np.zeros([chunk_size, shfit_fsmn], dtype=dtype)
				ones_3_right = np.ones([chunk_size, chunk_size], dtype=dtype)
				ones_3 = np.concatenate([zeros_3_left, ones_3_right], axis=1)

				zeros_remain_num = max(chunk_num - 1 - chunk_ids, 0)
				zeros_remain = np.zeros([chunk_size, zeros_remain_num*chunk_size_pad_shift], dtype=dtype)

				ones2_bottom = np.concatenate([zeros_2, ones_2, ones_3, zeros_remain], axis=1)
				mask_att_chunk_encoder_cur = np.concatenate([zeros_1_top, ones2_bottom], axis=0)
				mask_att_chunk_encoder = np.concatenate([mask_att_chunk_encoder, mask_att_chunk_encoder_cur], axis=0)


				# decoder fsmn_shift_att_mask
				zeros_1 = np.zeros([shfit_fsmn, 1])
				ones_1 = np.ones([chunk_size, 1])
				mask_shift_att_chunk_decoder_cur = np.concatenate([zeros_1, ones_1], axis=0)
				mask_shift_att_chunk_decoder = np.concatenate(
					[mask_shift_att_chunk_decoder, mask_shift_att_chunk_decoder_cur], axis=0)

			self.x_add_mask = x_add_mask[:x_len_chunk_max, :x_len_max+pad_left]
			self.x_len_chunk = x_len_chunk
			self.x_rm_mask = x_rm_mask[:x_len_max, :x_len_chunk_max]
			self.x_len = x_len
			self.mask_shfit_chunk = mask_shfit_chunk[:x_len_chunk_max, :]
			self.mask_chunk_predictor = mask_chunk_predictor[:x_len_chunk_max, :]
			self.mask_att_chunk_encoder = mask_att_chunk_encoder[:x_len_chunk_max, :x_len_chunk_max]
			self.mask_shift_att_chunk_decoder = mask_shift_att_chunk_decoder[:x_len_chunk_max, :]
			self.chunk_outs = (self.x_add_mask,
		        self.x_len_chunk,
		        self.x_rm_mask,
		        self.x_len,
		        self.mask_shfit_chunk,
		        self.mask_chunk_predictor,
		        self.mask_att_chunk_encoder,
		        self.mask_shift_att_chunk_decoder)

		return self.chunk_outs


	def split_chunk(self, x, x_len, chunk_outs):
		"""
		:param x: (b, t, d)
		:param x_length: (b)
		:param ind: int
		:return:
		"""
		x = x[:, :x_len.max(), :]
		b, t, d = x.size()
		x_len_mask = (~make_pad_mask(x_len, maxlen=t)).to(
			x.device)
		x *= x_len_mask[:, :, None]

		x_add_mask = self.get_x_add_mask(chunk_outs, x.device, dtype=x.dtype)
		x_len_chunk = self.get_x_len_chunk(chunk_outs, x_len.device, dtype=x_len.dtype)
		pad = (0, 0, self.pad_left_cur, 0)
		x = F.pad(x, pad, "constant", 0.0)
		b, t, d = x.size()
		x = torch.transpose(x, 1, 0)
		x = torch.reshape(x, [t, -1])
		x_chunk = torch.mm(x_add_mask, x)
		x_chunk = torch.reshape(x_chunk, [-1, b, d]).transpose(1, 0)

		return x_chunk, x_len_chunk

	def remove_chunk(self, x_chunk, x_len_chunk, chunk_outs):
		x_chunk = x_chunk[:, :x_len_chunk.max(), :]
		b, t, d = x_chunk.size()
		x_len_chunk_mask = (~make_pad_mask(x_len_chunk, maxlen=t)).to(
			x_chunk.device)
		x_chunk *= x_len_chunk_mask[:, :, None]

		x_rm_mask = self.get_x_rm_mask(chunk_outs, x_chunk.device, dtype=x_chunk.dtype)
		x_len = self.get_x_len(chunk_outs, x_len_chunk.device, dtype=x_len_chunk.dtype)
		x_chunk = torch.transpose(x_chunk, 1, 0)
		x_chunk = torch.reshape(x_chunk, [t, -1])
		x = torch.mm(x_rm_mask, x_chunk)
		x = torch.reshape(x, [-1, b, d]).transpose(1, 0)

		return x, x_len

	def get_x_add_mask(self, chunk_outs=None, device='cpu', idx=0, dtype=torch.float32):
		with torch.no_grad():
			x = chunk_outs[idx] if chunk_outs is not None else  self.chunk_outs[idx]
			x = torch.from_numpy(x).type(dtype).to(device)
		return x

	def get_x_len_chunk(self, chunk_outs=None, device='cpu', idx=1, dtype=torch.float32):
		with torch.no_grad():
			x = chunk_outs[idx] if chunk_outs is not None else  self.chunk_outs[idx]
			x = torch.from_numpy(x).type(dtype).to(device)
		return x


	def get_x_rm_mask(self, chunk_outs=None, device='cpu', idx=2, dtype=torch.float32):
		with torch.no_grad():
			x = chunk_outs[idx] if chunk_outs is not None else  self.chunk_outs[idx]
			x = torch.from_numpy(x).type(dtype).to(device)
		return x

	def get_x_len(self, chunk_outs=None, device='cpu', idx=3, dtype=torch.float32):
		with torch.no_grad():
			x = chunk_outs[idx] if chunk_outs is not None else  self.chunk_outs[idx]
			x = torch.from_numpy(x).type(dtype).to(device)
		return x


	def get_mask_shfit_chunk(self, chunk_outs=None, device='cpu', batch_size=1, num_units=1, idx=4, dtype=torch.float32):
		with torch.no_grad():
			x = chunk_outs[idx] if chunk_outs is not None else  self.chunk_outs[idx]
			x = np.tile(x[None, :, :, ], [batch_size, 1, num_units])
			x = torch.from_numpy(x).type(dtype).to(device)
		return x

	def get_mask_chunk_predictor(self, chunk_outs=None, device='cpu', batch_size=1, num_units=1, idx=5, dtype=torch.float32):
		with torch.no_grad():
			x = chunk_outs[idx] if chunk_outs is not None else  self.chunk_outs[idx]
			x = np.tile(x[None, :, :, ], [batch_size, 1, num_units])
			x = torch.from_numpy(x).type(dtype).to(device)
		return x

	def get_mask_att_chunk_encoder(self, chunk_outs=None, device='cpu', batch_size=1, idx=6, dtype=torch.float32):
		with torch.no_grad():
			x = chunk_outs[idx] if chunk_outs is not None else  self.chunk_outs[idx]
			x = np.tile(x[None, :, :, ], [batch_size, 1, 1])
			x = torch.from_numpy(x).type(dtype).to(device)
		return x

	def get_mask_shift_att_chunk_decoder(self, chunk_outs=None, device='cpu', batch_size=1, idx=7, dtype=torch.float32):
		with torch.no_grad():
			x = chunk_outs[idx] if chunk_outs is not None else  self.chunk_outs[idx]
			x = np.tile(x[None, None, :, 0], [batch_size, 1, 1])
			x = torch.from_numpy(x).type(dtype).to(device)
		return x



def build_scama_mask_for_cross_attention_decoder(
							  predictor_alignments: torch.Tensor,
                              encoder_sequence_length: torch.Tensor,
                              chunk_size: int = 5,
                              encoder_chunk_size: int = 5,
                              attention_chunk_center_bias: int = 0,
                              attention_chunk_size: int = 1,
                              attention_chunk_type: str = 'chunk',
                              step=None,
							  predictor_mask_chunk_hopping: torch.Tensor = None,
							  decoder_att_look_back_factor: int = 1,
							  mask_shift_att_chunk_decoder: torch.Tensor = None,
							  target_length: torch.Tensor = None,
							  is_training=True,
                              dtype: torch.dtype = torch.float32):
	with torch.no_grad():
		device = predictor_alignments.device
		batch_size, chunk_num = predictor_alignments.size()
		maximum_encoder_length = encoder_sequence_length.max().item()
		int_type = predictor_alignments.dtype
		if not is_training:
			target_length = predictor_alignments.sum(dim=-1).type(encoder_sequence_length.dtype)
		maximum_target_length = target_length.max()
		predictor_alignments_cumsum = torch.cumsum(predictor_alignments, dim=1)
		predictor_alignments_cumsum = predictor_alignments_cumsum[:, None, :].repeat(1, maximum_target_length, 1)
	
	
		index = torch.ones([batch_size, maximum_target_length], dtype=int_type).to(device)
		index = torch.cumsum(index, dim=1)
		index = index[:, :, None].repeat(1, 1, chunk_num)
	
		index_div = torch.floor(torch.divide(predictor_alignments_cumsum, index)).type(int_type)
		index_div_bool_zeros = index_div == 0
		index_div_bool_zeros_count = torch.sum(index_div_bool_zeros.type(int_type), dim=-1) + 1
	
		index_div_bool_zeros_count = torch.clip(index_div_bool_zeros_count, min=1, max=chunk_num)
	
		index_div_bool_zeros_count *= chunk_size
		index_div_bool_zeros_count += attention_chunk_center_bias
		index_div_bool_zeros_count = torch.clip(index_div_bool_zeros_count-1, min=0, max=maximum_encoder_length)
		index_div_bool_zeros_count_ori = index_div_bool_zeros_count
	
		index_div_bool_zeros_count = (torch.floor(index_div_bool_zeros_count / encoder_chunk_size)+1)*encoder_chunk_size
		max_len_chunk = math.ceil(maximum_encoder_length / encoder_chunk_size) * encoder_chunk_size
	
		mask_flip, mask_flip2 = None, None
		if attention_chunk_size is not None:
			index_div_bool_zeros_count_beg = index_div_bool_zeros_count - attention_chunk_size
			index_div_bool_zeros_count_beg = torch.clip(index_div_bool_zeros_count_beg, 0, max_len_chunk)
			index_div_bool_zeros_count_beg_mask = sequence_mask(index_div_bool_zeros_count_beg, maxlen=max_len_chunk, dtype=int_type, device=device)
			mask_flip = 1 - index_div_bool_zeros_count_beg_mask
			attention_chunk_size2 = attention_chunk_size * (decoder_att_look_back_factor+1)
			index_div_bool_zeros_count_beg = index_div_bool_zeros_count - attention_chunk_size2
	
			index_div_bool_zeros_count_beg = torch.clip(index_div_bool_zeros_count_beg, 0, max_len_chunk)
			index_div_bool_zeros_count_beg_mask = sequence_mask(index_div_bool_zeros_count_beg, maxlen=max_len_chunk, dtype=int_type, device=device)
			mask_flip2 = 1 - index_div_bool_zeros_count_beg_mask
	
		mask = sequence_mask(index_div_bool_zeros_count, maxlen=max_len_chunk, dtype=dtype, device=device)
	
		if predictor_mask_chunk_hopping is not None:
				b, k, t = mask.size()
				predictor_mask_chunk_hopping = predictor_mask_chunk_hopping[:, None, :, 0].repeat(1, k, 1)
	
				mask_mask_flip = mask
				if mask_flip is not None:
						mask_mask_flip = mask_flip * mask
	
				def _fn():
						mask_sliced = mask[:b, :k, encoder_chunk_size:t]
						zero_pad_right = torch.zeros([b, k, encoder_chunk_size], dtype=mask_sliced.dtype).to(device)
						mask_sliced = torch.cat([mask_sliced, zero_pad_right], dim=2)
						_, _, tt = predictor_mask_chunk_hopping.size()
						pad_right_p = max_len_chunk - tt
						predictor_mask_chunk_hopping_pad = torch.nn.functional.pad(predictor_mask_chunk_hopping, [0, pad_right_p], "constant", 0)
						masked = mask_sliced * predictor_mask_chunk_hopping_pad
	
						mask_true = mask_mask_flip + masked
						return mask_true
	
				mask = _fn() if t > chunk_size else mask_mask_flip
	
	
	
		if mask_flip2 is not None:
			mask *= mask_flip2
	
		mask_target = sequence_mask(target_length, maxlen=maximum_target_length, dtype=mask.dtype, device=device)
		mask = mask[:, :maximum_target_length, :] * mask_target[:, :, None]
	
	
	
		mask_len = sequence_mask(encoder_sequence_length, maxlen=maximum_encoder_length, dtype=mask.dtype, device=device)
		mask = mask[:, :, :maximum_encoder_length] * mask_len[:, None, :]
	
	
	
	
		if attention_chunk_type == 'full':
			mask = torch.ones_like(mask).to(device)
		if mask_shift_att_chunk_decoder is not None:
			mask = mask * mask_shift_att_chunk_decoder
		mask = mask[:, :maximum_target_length, :maximum_encoder_length].type(dtype).to(device)

	return mask




import os
import io
import time
import logging
import wave
from pathlib import Path

import numpy as np
import s3tokenizer
import onnxruntime
import torch
import torchaudio
import torchaudio.compliance.kaldi as kaldi
from flashcosyvoice.modules.hifigan import HiFTGenerator
from flashcosyvoice.utils.audio import mel_spectrogram
from hyperpyyaml import load_hyperpyyaml

def fade_in_out(fade_in_mel:torch.Tensor, fade_out_mel:torch.Tensor, window:torch.Tensor):
    """perform fade_in_out in tensor style
    """
    mel_overlap_len = int(window.shape[0] / 2)
    fade_in_mel = fade_in_mel.clone()
    fade_in_mel[..., :mel_overlap_len] = \
        fade_in_mel[..., :mel_overlap_len] * window[:mel_overlap_len] + \
        fade_out_mel[..., -mel_overlap_len:] * window[mel_overlap_len:]
    return fade_in_mel


class Token2wav():
    CHUNK_SIZE = 25
    WARMUP_TOKENS = [1493, 4299, 4218, 2049, 528, 2752, 4850, 4569, 4575, 6372, 2127, 4068, 2312, 4993, 4769, 2300, 226, 2175, 2160, 2152, 6311, 6065, 4859, 5102, 4615, 6534, 6426, 1763, 2249, 2209, 5938, 1725, 6048, 3816, 6058, 958, 63, 4460, 5914, 2379, 735, 5319, 4593, 2328, 890, 35, 751, 1483, 1484, 1483, 2112, 303, 4753, 2301, 5507, 5588, 5261, 5744, 5501, 2341, 2001, 2252, 2344, 1860, 2031, 414, 4366, 4366, 6059, 5300, 4814, 5092, 5100, 1923, 3054, 4320, 4296, 2148, 4371, 5831, 5084, 5027, 4946, 4946, 2678, 575, 575, 521, 518, 638, 1367, 2804, 3402, 4299]

    def __init__(self, model_path, float16=False, prompt_wav:str="assets/default_female.wav", **kwargs):
        self.float16 = float16

        logging.info(f"init token2wav ...")
        self.audio_tokenizer = s3tokenizer.load_model(f"{model_path}/speech_tokenizer_v2_25hz.onnx")

        self.device = "cpu"
        if torch.cuda.is_available():
            self.device = "cuda"
        self.audio_tokenizer.to(self.device).eval()

        option = onnxruntime.SessionOptions()
        option.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL
        option.intra_op_num_threads = 1
        self.spk_model = onnxruntime.InferenceSession(f"{model_path}/campplus.onnx", sess_options=option, providers=["CPUExecutionProvider"])

        with open(f"{model_path}/flow.yaml", "r") as f:
            configs = load_hyperpyyaml(f)
            self.flow = configs['flow']
        if float16:
            self.flow.half()
        self.flow.load_state_dict(torch.load(f"{model_path}/flow.pt", map_location="cpu", weights_only=True), strict=True)
        self.flow.to(self.device).eval()

        if torch.cuda.is_available() and kwargs.get("cuda_graph", False):
            logging.info(f"move token2wav flow to cuda and scatter_cuda_graph ...")
            self.flow.scatter_cuda_graph(True)

        self.hift = HiFTGenerator()
        hift_state_dict = {k.replace('generator.', ''): v for k, v in torch.load(f"{model_path}/hift.pt", map_location="cpu", weights_only=True).items()}
        self.hift.load_state_dict(hift_state_dict, strict=True)
        self.hift.to(self.device).eval()

        self.cache = {}
        self.stream_cache = None

        # stream conf
        self.mel_cache_len = 8  # hard-coded, 160ms
        self.source_cache_len = int(self.mel_cache_len * 480)   # 50hz mel -> 24kHz wave
        self.speech_window = torch.from_numpy(np.hamming(2 * self.source_cache_len)).to(self.device)

        # hifigan cache
        self.hift_cache_dict = {}

        self._verbose = kwargs.get("verbose", False)

        if kwargs.get("warmup_cn", 0) > 0:
            self.warmup(prompt_wav, kwargs.get("warmup_cn"))


    def _prepare_prompt(self, prompt_wav):
        audio = s3tokenizer.load_audio(prompt_wav, sr=16000)  # [T]
        mels = s3tokenizer.log_mel_spectrogram(audio)
        mels, mels_lens = s3tokenizer.padding([mels])
        prompt_speech_tokens, prompt_speech_tokens_lens = self.audio_tokenizer.quantize(mels.to(self.device), mels_lens.to(self.device))

        spk_feat = kaldi.fbank(audio.unsqueeze(0), num_mel_bins=80, dither=0, sample_frequency=16000)
        spk_feat = spk_feat - spk_feat.mean(dim=0, keepdim=True)
        spk_emb = torch.tensor(self.spk_model.run(
            None, {self.spk_model.get_inputs()[0].name: spk_feat.unsqueeze(dim=0).cpu().numpy()}
        )[0], device=self.device)

        audio, sample_rate = torchaudio.load(prompt_wav, backend='soundfile')
        audio = audio.mean(dim=0, keepdim=True)  # [1, T]
        if sample_rate != 24000:
            audio = torchaudio.transforms.Resample(orig_freq=sample_rate, new_freq=24000)(audio)
        prompt_mel = mel_spectrogram(audio).transpose(1, 2).squeeze(0)  # [T, num_mels]
        prompt_mels = prompt_mel.unsqueeze(0).to(self.device)
        prompt_mels_lens = torch.tensor([prompt_mels.shape[1]], dtype=torch.int32, device=self.device)
        prompt_mels = torch.nn.functional.pad(prompt_mels, (0, 0, 0, prompt_speech_tokens.shape[1] * self.flow.up_rate - prompt_mels.shape[1]), mode='replicate')
        return prompt_speech_tokens, prompt_speech_tokens_lens, spk_emb, prompt_mels, prompt_mels_lens

    def __call__(self, generated_speech_tokens, prompt_wav):
        if prompt_wav not in self.cache:
            self.cache[prompt_wav] = self._prepare_prompt(prompt_wav)
        prompt_speech_tokens, prompt_speech_tokens_lens, spk_emb, prompt_mels, prompt_mels_lens = self.cache[prompt_wav]

        generated_speech_tokens = torch.tensor([generated_speech_tokens], dtype=torch.int32, device=self.device)
        generated_speech_tokens_lens = torch.tensor([generated_speech_tokens.shape[1]], dtype=torch.int32, device=self.device)

        with torch.amp.autocast(self.device, dtype=torch.float16 if self.float16 else torch.float32):
            mel = self.flow.inference(generated_speech_tokens, generated_speech_tokens_lens,
                prompt_speech_tokens, prompt_speech_tokens_lens,
                prompt_mels, prompt_mels_lens, spk_emb, 10)

        wav, _ = self.hift(speech_feat=mel)
        output = io.BytesIO()
        torchaudio.save(output, wav.cpu(), sample_rate=24000, format='wav')

        return output.getvalue()

    def set_stream_cache(self, prompt_wav):
        logging.info(f"set_stream_cache {prompt_wav}")
        if prompt_wav not in self.cache:
            self.cache[prompt_wav] = self._prepare_prompt(prompt_wav)
        prompt_speech_tokens, prompt_speech_tokens_lens, spk_emb, prompt_mels, prompt_mels_lens = self.cache[prompt_wav]
        self.stream_cache = self.flow.setup_cache(
            torch.cat([prompt_speech_tokens, prompt_speech_tokens[:, :3]], dim=1),
            prompt_mels, spk_emb, n_timesteps=10)
        # hift cache
        self.hift_cache_dict = dict(
            mel = torch.zeros(1, prompt_mels.shape[2], 0, device=self.device), 
            source = torch.zeros(1, 1, 0, device=self.device),
            speech = torch.zeros(1, 0, device=self.device),
        )

    def stream(self, generated_speech_tokens, prompt_wav, last_chunk=False):
        if prompt_wav not in self.cache:
            self.cache[prompt_wav] = self._prepare_prompt(prompt_wav)
        prompt_speech_tokens, prompt_speech_tokens_lens, spk_emb, prompt_mels, prompt_mels_lens = self.cache[prompt_wav]

        generated_speech_tokens = torch.tensor([generated_speech_tokens], dtype=torch.int32, device=self.device)
        generated_speech_tokens_lens = torch.tensor([generated_speech_tokens.shape[1]], dtype=torch.int32, device=self.device)

        if self.stream_cache is None:
            # raise ValueError("stream_cache is not set")
            logging.warning("stream_cache is not set, then set it")
            self.set_stream_cache(prompt_wav)

        if self._verbose is True:
            for k, v in self.stream_cache.items():
                print(f"{k}: {v.shape}")
        
        with torch.amp.autocast(self.device, dtype=torch.float16 if self.float16 else torch.float32):
            start = time.time()
            chunk_mel, self.stream_cache = self.flow.inference_chunk(
                token=generated_speech_tokens,
                spk=spk_emb,
                cache=self.stream_cache,
                last_chunk=last_chunk,
                n_timesteps=10,
            )
            elapsed = time.time() - start
            if self._verbose is True:
                print(f"flow inference_chunk time: {elapsed:.3f}s")
        if self.stream_cache['estimator_att_cache'].shape[4] > (prompt_mels.shape[1] + 100):
            self.stream_cache['estimator_att_cache'] = torch.cat([
                self.stream_cache['estimator_att_cache'][:, :, :, :, :prompt_mels.shape[1]],
                self.stream_cache['estimator_att_cache'][:, :, :, :, -100:],
            ], dim=4)
        
        # vocoder cache
        hift_cache_mel = self.hift_cache_dict['mel']
        hift_cache_source = self.hift_cache_dict['source']
        hift_cache_speech = self.hift_cache_dict['speech']
        mel = torch.concat([hift_cache_mel, chunk_mel], dim=2)

        start = time.time()
        speech, source = self.hift(mel, hift_cache_source)
        elapsed = time.time() - start
        if self._verbose is True:
            print(f"hift inference_chunk time: {elapsed:.3f}s")

        # overlap speech smooth
        if hift_cache_speech.shape[-1] > 0:
            speech = fade_in_out(speech, hift_cache_speech, self.speech_window)

        # update vocoder cache
        self.hift_cache_dict = dict(
            mel = mel[..., -self.mel_cache_len:].clone().detach(),
            source = source[:, :, -self.source_cache_len:].clone().detach(),
            speech = speech[:, -self.source_cache_len:].clone().detach(),
        )
        if not last_chunk:
            speech = speech[:, :-self.source_cache_len]

        wav_np = speech.cpu().numpy()
        # Clip to [-1, 1] to avoid overflow, then scale to int16
        wav_np = np.clip(wav_np, -1.0, 1.0)
        wav_int16 = (wav_np * 32767.0).astype('<i2')  # 16-bit little-endian PCM
        pcm_bytes = wav_int16.tobytes()
        return pcm_bytes
    
    def warmup(self, prompt_wav, warmup_cn:int=2):
        if warmup_cn > 0:
            for i in range(warmup_cn):
                start = time.time()
                self.stream(self.WARMUP_TOKENS[:self.CHUNK_SIZE + self.flow.pre_lookahead_len], prompt_wav=prompt_wav) # Warm up
                logging.info(f"Token2wav warmup {i=} done in {time.time() - start:.3f}s")
            # NOTE: reset stream cache
            self.set_stream_cache(prompt_wav)


def test_stream(token2wav: Token2wav):
    prompt_wav = "assets/default_male.wav"
    #prompt_wav = "assets/default_female.wav"
    token2wav.set_stream_cache(prompt_wav)
    token2wav.warmup(prompt_wav, warmup_cn=2)

    buffer = []
    pcm = b""

    output_file = Path("output_chunks_stream_tts.wav")
    output_file.unlink(missing_ok=True)

    for audio_token_id in token2wav.WARMUP_TOKENS:
        buffer.append(audio_token_id)
        if len(buffer) >= token2wav.CHUNK_SIZE + token2wav.flow.pre_lookahead_len:
            start = time.time()
            output = token2wav.stream(
                buffer[: token2wav.CHUNK_SIZE + token2wav.flow.pre_lookahead_len],
                prompt_wav=prompt_wav,
                last_chunk=False,
            )
            print(len(buffer), len(output), output[:50],time.time()-start)
            pcm += output
            buffer = buffer[token2wav.CHUNK_SIZE:]

    if len(buffer) > 0:
        start = time.time()
        output = token2wav.stream(buffer, prompt_wav=prompt_wav, last_chunk=True)
        print("last_chunk", len(buffer), len(output), output[:50], time.time()-start)
        pcm += output
    with wave.open(str(output_file), "wb") as wf:
        wf.setnchannels(1)
        wf.setsampwidth(2)
        wf.setframerate(24000)
        wf.writeframes(pcm)

def test_token2wav(token2wav:Token2wav):
    audio = token2wav(token2wav.WARMUP_TOKENS, 'assets/default_male.wav')
    with open('give_me_a_brief_introduction_to_the_great_wall.wav', 'wb') as f:
        f.write(audio)

"""
python -m token2wav
TEST_FUNC=test_stream python -m token2wav
"""
if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)

    # huggingface-cli download stepfun-ai/Step-Audio-2-mini --include token2wav/ --local-dir Step-Audio-2-mini
    token2wav = Token2wav('Step-Audio-2-mini/token2wav')

    test_func = os.getenv("TEST_FUNC","test_token2wav")
    globals()[test_func](token2wav)



import logging
import time
import os
import shutil
import sys
import json
import subprocess as sp
from typing import Dict, Optional, List


import librosa
import numpy as np
from tqdm import tqdm


from achatbot.common.time_utils import timeit
from achatbot.common.types import RESOURCES_DIR, MODELS_DIR, ASSETS_DIR
from achatbot.common.factory import EngineClass
from achatbot.modules.avatar.interface import IFaceAvatar
from achatbot.types.avatar.lam_audio2expression import LAMAudio2ExpressionAvatarArgs


try:
    cur_dir = os.path.dirname(__file__)
    deps_dir = os.path.join(cur_dir, "../../../deps/LAM_Audio2Expression")
    if bool(os.getenv("ACHATBOT_PKG", "")) is True:
        deps_dir = os.path.join(cur_dir, "../../LAM_Audio2Expression")
    sys.path.insert(0, deps_dir)

    import torch
    from achatbot.LAM_Audio2Expression.engines.defaults import (
        default_config_parser,
        default_setup,
    )
    from achatbot.LAM_Audio2Expression.engines.infer import (
        INFER,
        Audio2ExpressionInfer,
        export_blendshape_animation,
    )
except ModuleNotFoundError as e:
    logging.error(f"Exception: {e}")
    logging.error(
        "In order to use Audio2Expression-avatar, you need to `pip install achatbot[lam_audio2expression_avatar]`. "
    )
    raise Exception(f"Missing module: {e}")


class LAMAudio2ExpressionAvatar(EngineClass):
    TAG = "lam_audio2expression_avatar"

    def __init__(self, **kwargs):
        super().__init__()
        self.args = LAMAudio2ExpressionAvatarArgs(**kwargs)
        logging.info(f"{self.args=}")
        self.infer: Audio2ExpressionInfer = None
        self.arkit_channels: List[str] = []

    def load(self):
        config_file = os.path.join(
            deps_dir,
            "configs",
            "lam_audio2exp_config_streaming.py",
        )
        wav2vec_config_file = os.path.join(
            deps_dir,
            "configs",
            "wav2vec2_config.json",
        )
        cfg = default_config_parser(
            config_file,
            {
                "weight": self.args.weight_path,
                "audio_sr": self.args.avatar_audio_sample_rate,
                "fps": self.args.fps,
                "model": {
                    "backbone": {
                        "pretrained_encoder_path": self.args.wav2vec_dir,
                        "wav2vec2_config_path": wav2vec_config_file,
                        "expression_dim": self.args.expression_dim,
                    }
                },
            },
        )
        self.cfg = default_setup(cfg)
        logging.info(f"{self.cfg=}")

        self.infer = INFER.build(dict(type=self.cfg.infer.type, cfg=self.cfg))
        self.infer.model.eval()
        arkit_channel_list_path = os.path.join(ASSETS_DIR, "arkit_face_channels.txt")
        self.arkit_channels.clear()
        with open(arkit_channel_list_path, "r") as f:
            for line in f:
                self.arkit_channels.append(line.strip())

        self.warm_up()

    def warm_up(self):
        t_start = time.monotonic()
        context: Optional[Dict] = None
        self.infer.infer_streaming_audio(
            context=context,
            audio=np.zeros([self.args.avatar_audio_sample_rate], dtype=np.float32),
            ssr=self.args.avatar_audio_sample_rate,
        )
        dur_warmup = time.monotonic() - t_start
        logging.info(f"LAM_Audio2Expression warmup finished in {dur_warmup * 1000} milliseconds.")

    def export_animation_json(self, bs_array, json_path=None):
        if json_path is not None:
            os.makedirs(os.path.dirname(json_path), exist_ok=True)
        animation_data = export_blendshape_animation(
            bs_array, json_path, self.arkit_channels, fps=30.0
        )
        return json.dumps(animation_data)


"""
wget https://virutalbuy-public.oss-cn-hangzhou.aliyuncs.com/share/aigc3d/data/LAM/LAM_audio2exp_streaming.tar -P ./models/LAM_audio2exp/
tar -xzvf ./models/LAM_audio2exp/LAM_audio2exp_streaming.tar -C ./models/LAM_audio2exp && rm ./models/LAM_audio2exp/LAM_audio2exp_streaming.tar
git clone --depth 1 https://www.modelscope.cn/AI-ModelScope/wav2vec2-base-960h.git ./models/facebook/wav2vec2-base-960h

python -m src.modules.avatar.lam_audio2expression

python -m src.modules.avatar.lam_audio2expression \
    --audio_path test/audio_files/asr_example_zh.wav \
    --save_json_path resources/avatar/lam_audio2expression/expression.json
"""

if __name__ == "__main__":
    import argparse
    import logging

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--weight_path",
        type=str,
        default=os.path.join(
            MODELS_DIR, "LAM_audio2exp/pretrained_models/lam_audio2exp_streaming.tar"
        ),
    )
    parser.add_argument(
        "--wav2vec_dir", type=str, default=os.path.join(MODELS_DIR, "facebook/wav2vec2-base-960h")
    )
    parser.add_argument("--audio_sample_rate", type=int, default=16000)
    parser.add_argument(
        "--audio_path",
        type=str,
        default=os.path.join(deps_dir, "assets/sample_audio/BarackObama.wav"),
    )
    parser.add_argument(
        "--save_json_path",
        type=str,
        default=os.path.join(RESOURCES_DIR, "avatar/lam_audio2expression/expression.json"),
    )
    args = parser.parse_args()
    log_format = "%(asctime)s - %(name)s - %(levelname)s - %(pathname)s:%(lineno)d - %(funcName)s - %(message)s"
    logging.basicConfig(level="INFO", format=log_format)

    audio, sample_rate = librosa.load(args.audio_path, sr=16000)
    print(f"{audio=} {sample_rate=} {audio.shape[0]=}")

    avatar = LAMAudio2ExpressionAvatar(
        **LAMAudio2ExpressionAvatarArgs(
            weight_path=args.weight_path,
            wav2vec_dir=args.wav2vec_dir,
            speaker_audio_sample_rate=sample_rate,
            avatar_audio_sample_rate=args.audio_sample_rate,
        ).__dict__
    )
    avatar.load()

    context = None
    input_num = (audio.shape[0] + 15999) // sample_rate
    gap = sample_rate
    all_exp = []
    for i in tqdm(range(input_num)):
        start = time.time()
        # infer streaming audio with 30 fps
        output, context = avatar.infer.infer_streaming_audio(
            audio[i * gap : (i + 1) * gap], sample_rate, context
        )
        end = time.time()
        print("Inference time {}".format(end - start))
        print(f"{output['expression'].shape=}")
        all_exp.append(output["expression"])

    all_exp = np.concatenate(all_exp, axis=0)
    print(f"{all_exp.shape=}")

    animation_json_str = avatar.export_animation_json(all_exp, args.save_json_path)
    # print(f"{animation_json_str=}")



import logging
from threading import Thread
from PIL import Image
from time import perf_counter

try:
    from transformers import AutoProcessor, TextIteratorStreamer, Gemma3nForConditionalGeneration
    import torch

except ModuleNotFoundError as e:
    logging.error(f"Exception: {e}")
    logging.error(
        "In order to use Gemma3n, you need to `pip install achatbot[llm_transformers_manual_vision_speech_gemma]`"
    )
    raise Exception(f"Missing module: {e}")


from src.common.utils.helper import get_device, print_model_params
from src.common.random import set_all_random_seed
from src.common.chat_history import ChatHistory
from src.common.session import Session
from src.types.speech.language import TO_LLM_LANGUAGE
from src.common.types import SessionCtx
from src.types.llm.transformers import TransformersLMArgs
from .base import TransformersBaseLLM


class TransformersManualVisionSpeechGemmaLM(TransformersBaseLLM):
    TAG = "llm_transformers_manual_vision_speech_gemma3n"

    def __init__(self, **args) -> None:
        self.args = TransformersLMArgs(**args)

        if self.args.lm_device_map:
            self._model = Gemma3nForConditionalGeneration.from_pretrained(
                self.args.lm_model_name_or_path,
                torch_dtype=torch.bfloat16,
                #!NOTE: https://github.com/huggingface/transformers/issues/20896
                # device_map for multi cpu/gpu with accelerate
                device_map=self.args.lm_device_map,
                # attn_implementation="flash_attention_2",
                trust_remote_code=True,
            ).eval()
        else:
            self._model = (
                Gemma3nForConditionalGeneration.from_pretrained(
                    self.args.lm_model_name_or_path,
                    torch_dtype=torch.bfloat16,
                    # attn_implementation="flash_attention_2",
                    trust_remote_code=True,
                )
                .eval()
                .to(self.args.lm_device)
            )

        logging.info(f"TransformersLMArgs: {self.args}")
        print_model_params(self._model, self.TAG)
        self._tokenizer = AutoProcessor.from_pretrained(
            self.args.lm_model_name_or_path, use_fast=True
        )

        self.session_chat_history = {}
        torch.set_float32_matmul_precision("high")

        self.warmup()

    def warmup(self):
        if self.args.warmup_steps <= 0:
            return
        dummy_input_text = self.args.warmup_prompt
        dummy_pil_image = Image.new("RGB", (100, 100), color="white")
        dummy_msgs = [
            {"type": "text", "text": dummy_input_text},
            {"type": "image", "image": dummy_pil_image},
        ]

        session = Session(**SessionCtx(f"{self.SELECTED_TAG}_warmup").__dict__)
        for i in range(self.args.warmup_steps):
            logging.info(f"{self.SELECTED_TAG} warmup: {i + 1}/{self.args.warmup_steps}")
            session.ctx.state["prompt"] = dummy_msgs
            for item in self.generate(session, max_new_tokens=128):
                pass

    def generate(self, session: Session, **kwargs):
        seed = kwargs.get("seed", self.args.lm_gen_seed)
        set_all_random_seed(seed)

        prompt = session.ctx.state["prompt"]
        assert len(prompt) > 0

        message = {"role": self.args.user_role, "content": prompt}
        self.add_chat_history(session, message)
        chat_history = self.get_session_chat_history(session.ctx.client_id)
        logging.info(f"{session.ctx.client_id} chat_history:{chat_history}")
        inputs = self._tokenizer.apply_chat_template(
            chat_history,
            add_generation_prompt=True,
            tokenize=True,
            return_dict=True,
            return_tensors="pt",
        ).to(
            self._model.device,
            dtype=torch.bfloat16,
        )

        for key, value in inputs.items():
            logging.info(f"{key}: {value.shape=}")
        # input_ids = inputs["input_ids"]
        # prompt = self._tokenizer.decode(input_ids[0])
        # logging.info(f"{prompt=}")

        streamer = TextIteratorStreamer(self._tokenizer, skip_prompt=True, skip_special_tokens=True)

        generation_kwargs = dict(
            **inputs,
            streamer=streamer,
            do_sample=True
            if kwargs.get("temperature", self.args.lm_gen_temperature) > 0
            else False,
            temperature=kwargs.get("temperature", self.args.lm_gen_temperature),
            top_k=kwargs.get("top_k", self.args.lm_gen_top_k),
            top_p=kwargs.get("top_p", self.args.lm_gen_top_p),
            repetition_penalty=kwargs.get(
                "repetition_penalty", self.args.lm_gen_repetition_penalty
            ),
            min_new_tokens=kwargs.get("min_new_tokens", self.args.lm_gen_min_new_tokens),
            max_new_tokens=kwargs.get("max_new_tokens", self.args.lm_gen_max_new_tokens),
            use_cache=True,
            cache_implementation=kwargs.get(
                "cache_implementation", self.args.lm_gen_cache_implementation
            ),
        )
        thread = Thread(target=self._model.generate, kwargs=generation_kwargs)
        thread.start()

        generated_text = ""
        start = perf_counter()
        times = []
        with torch.inference_mode():
            for new_text in streamer:
                times.append(perf_counter() - start)
                generated_text += new_text
                yield {"text": new_text.replace("*", "")}
                start = perf_counter()
        logging.info(f"{generated_text=} TTFT: {times[0]:.4f}s total time: {sum(times):.4f}s")
        torch.cuda.empty_cache()
        self.session_chat_history[session.ctx.client_id].append(
            {
                "role": self.args.assistant_role,
                "content": [{"type": "text", "text": generated_text}],
            }
        )


class TransformersManualVisionGemma3nLM(TransformersManualVisionSpeechGemmaLM):
    TAG = "llm_transformers_manual_vision_gemma3n"

    def generate(self, session: Session, **kwargs):
        for item in super().generate(session, **kwargs):
            yield item["text"]


class TransformersManualAudioGemma3nLM(TransformersManualVisionSpeechGemmaLM):
    """
    audio understanding

    - speech -> text
    """

    TAG = [
        "llm_transformers_manual_gemma3n_audio_asr",
        "llm_transformers_manual_gemma3n_audio_translation",
    ]

    def generate(self, session: Session, **kwargs):
        for item in super().generate(session, **kwargs):
            text = item.pop("text", "")
            if text == "":
                continue
            yield text



from dataclasses import dataclass, field

from achatbot.types.llm.transformers import TransformersLMArgs


@dataclass
class KimiAudioDeTokenizerArgs:
    device: str = None
    look_ahead_tokens: int = 12
    max_prompt_chunk: int = 10  # 10 * 3 = 30s
    max_kv_cache_tokens: int = 900
    use_cfg: bool = False
    use_cfg_rescale: bool = True
    cfg_init: float = 1.5
    cfg_scale: float = 7.5
    cfg_schedule: str = "linear"  # no use


@dataclass
class KimiAudioTransformersVoiceLMArgs(TransformersLMArgs):
    """
    text+voice(audio+speech) lm args + double heads(text|audio) sampling  + code2wav(Audio-DeTokenizer)(dit cfm + bigvgan vocoder) args
    """

    is_load_detokenizer: bool = True
    code2wav_args: dict = field(default_factory=lambda: KimiAudioDeTokenizerArgs().__dict__)



import os
import json
import subprocess
import importlib
import uuid
from pathlib import Path
from typing import Dict

import modal
import modal.experimental

from app import model_dir, MODEL_VOL, comfyui_out_dir, COMFYUI_OUT_VOL


# if use nightly, git pull comfyui repo
VERSION = os.getenv("VERSION", "0.3.75")
IMAGE_GPU = os.getenv("IMAGE_GPU", "L4")
MODEL_NAME = os.getenv("MODEL_NAME", "flux1_schnell_fp8")

APP_DIR = Path(__file__).parent / "app"
image = (  # build up a Modal Image to run ComfyUI, step by step
    modal.Image.debian_slim(  # start from basic Linux with Python
        python_version="3.11"
    )
    .apt_install("git")  # install git to clone ComfyUI
    .uv_pip_install("fastapi[standard]==0.115.4")  # install web dependencies
    # https://github.com/Comfy-Org/comfy-cli (manage comfyui and model)
    .uv_pip_install("comfy-cli==1.5.3")  # install comfy-cli
    .run_commands(  # use comfy-cli to install ComfyUI and its dependencies
        # https://github.com/comfyanonymous/ComfyUI
        f"comfy --skip-prompt install --fast-deps --nvidia --version {VERSION}"
    )
    .run_commands(
        "which comfy",
        "comfy --help",
    )
    # copy the app path to the container.
    .add_local_dir(APP_DIR, f"/root/app", copy=True)
    .env(
        {
            "MODEL_NAME": MODEL_NAME,
        }
    )
)


# https://docs.comfy.org/installation/manual_install#example-structure
def link_comfyui_dir():
    try:
        MODEL_NAME = os.getenv("MODEL_NAME")
        module = importlib.import_module(f"app.{MODEL_NAME}")
        print(module)
        func = getattr(module, "link_comfyui_dir")
        func()
    except ImportError as e:
        print(e)
    except Exception as e:
        print(e)


image = image.run_function(
    link_comfyui_dir,
    volumes={model_dir: MODEL_VOL},
)


app = modal.App(name="comfyui", image=image)


@app.function(
    max_containers=1,  # limit interactive session to 1 container
    gpu=IMAGE_GPU,
    volumes={
        model_dir: MODEL_VOL,
        comfyui_out_dir: COMFYUI_OUT_VOL,
    },
)
@modal.concurrent(
    max_inputs=10
)  # required for UI startup process which runs several API calls concurrently
@modal.web_server(8000, startup_timeout=60)
def ui():
    subprocess.Popen("comfy launch -- --listen 0.0.0.0 --port 8000", shell=True)


"""
modal serve src/comfyui/ui.py 
IMAGE_GPU=L40S modal serve src/comfyui/ui.py 

MODEL_NAME=image_z_image_turbo modal serve src/comfyui/ui.py 
MODEL_NAME=image_z_image_turbo IMAGE_GPU=L40S modal serve src/comfyui/ui.py 

MODEL_NAME=image_flux2 IMAGE_GPU=L40S modal serve src/comfyui/ui.py 

# hunyuan video 1.5
MODEL_NAME=video_hunyuan_video_1.5_720p_t2v IMAGE_GPU=L40S modal serve src/comfyui/ui.py 
"""



import os
import asyncio
import json
import re
from collections.abc import AsyncGenerator, Callable, Generator
from pathlib import Path
from typing import Literal

from jinja2 import Template
from mcp.types import CallToolResult
from dotenv import load_dotenv

from .mcp import call_mcp_tool, get_mcp_tool_prompt

load_dotenv()

GOOGLE_API_KEY = os.environ.get("GOOGLE_API_KEY")

dir_path = Path(__file__).parent

with Path(dir_path / "tpl" / "decide.jinja").open("r") as f:
    decide_template = Template(f.read())

with Path(dir_path / "tpl" / "tool.jinja").open("r") as f:
    tool_template = Template(f.read())

with Path(dir_path / "tpl" / "called_tools_history.jinja").open("r") as f:
    called_tools_history_template = Template(f.read())


def stream_llm(prompt: str) -> Generator[str, None]:
    """Stream LLM response.

    Args:
        prompt (str): The prompt to send to the LLM.

    Returns:
        Generator[str, None, None]: A generator of the LLM response.
    !TODO: use openai / local inference engine (vLLM/SGLang/TensorRT-LLM wrapped by achatbot)
    """
    from google import genai

    client = genai.Client(vertexai=False, api_key=GOOGLE_API_KEY)
    for chunk in client.models.generate_content_stream(
        model="gemini-2.5-flash-lite",
        contents=prompt,
    ):
        if chunk.text:
            yield chunk.text


class Agent:
    """Agent for interacting with the Google Gemini LLM in different modes."""

    def __init__(
        self,
        mode: Literal["complete", "stream"] = "stream",
        token_stream_callback: Callable[[str], None] | None = None,
        mcp_url: str | None = None,
    ):
        self.mode = mode
        self.token_stream_callback = token_stream_callback
        self.mcp_url = mcp_url

    def call_llm(self, prompt: str) -> Generator[str, None]:
        """Call the LLM with the given prompt and return a generator of responses.

        Args:
            prompt (str): The prompt to send to the LLM.

        Returns:
            Generator[str, None]: A generator yielding the LLM's response.
        """
        return stream_llm(prompt)

    async def decide(
        self, question: str, called_tools: list[dict] | None = None
    ) -> Generator[str, None]:
        """Decide which tool to use to answer the question.

        Args:
            question (str): The question to answer.
            called_tools (list[dict]): The tools that have been called.
        """
        if self.mcp_url is None:
            return self.call_llm(question)
        tool_prompt = await get_mcp_tool_prompt(self.mcp_url)
        if called_tools:
            called_tools_prompt = called_tools_history_template.render(called_tools=called_tools)
        else:
            called_tools_prompt = ""

        prompt = decide_template.render(
            question=question,
            tool_prompt=tool_prompt,
            called_tools=called_tools_prompt,
        )
        return self.call_llm(prompt)

    def extract_tools(self, response: str) -> list[dict]:
        """Extract the tools from the response.

        Args:
            response (str): The response from the LLM.
        """
        pattern = r"```json\n(.*?)\n```"
        match = re.search(pattern, response, re.DOTALL)
        if match:
            return json.loads(match.group(1))
        return []

    async def call_tool(self, tools: list[dict]) -> list[CallToolResult]:
        """Call the tool.

        Args:
            tools (list[dict]): The tools to call.
        """
        return await asyncio.gather(
            *[call_mcp_tool(self.mcp_url, tool["name"], tool["arguments"]) for tool in tools]
        )

    async def stream(self, question: str) -> AsyncGenerator[str]:
        """Stream the process of answering a question, possibly involving tool calls.

        Args:
            question (str): The question to answer.

        Yields:
            dict: Streaming output, including intermediate steps and final result.
        """
        called_tools = []
        for i in range(10):
            yield {
                "is_task_complete": False,
                "require_user_input": False,
                "content": f"Step {i}",
            }

            response = ""
            for chunk in await self.decide(question, called_tools):
                response += chunk
                yield {
                    "is_task_complete": False,
                    "require_user_input": False,
                    "content": chunk,
                }
            tools = self.extract_tools(response)
            if not tools:
                break
            results = await self.call_tool(tools)  # gather concurrency

            called_tools += [
                {
                    "tool": tool["name"],
                    "arguments": tool["arguments"],
                    "isError": result.isError,
                    "result": result.content[0].text,
                }
                for tool, result in zip(tools, results, strict=True)
            ]
            called_tools_history = called_tools_history_template.render(
                called_tools=called_tools, question=question
            )
            yield {
                "is_task_complete": False,
                "require_user_input": False,
                "content": called_tools_history,
            }

        yield {
            "is_task_complete": True,
            "require_user_input": False,
            #"content": "Task completed",
            "content": response
        }


"""
python -m a2a_server.cmd.mcp_github.agent
"""
if __name__ == "__main__":
    agent = Agent(
        token_stream_callback=lambda token: print(token, end="", flush=True),
        mcp_url="https://gitmcp.io/ai-bot-pro/achatbot",
    )

    async def main():
        """Main function."""
        async for chunk in agent.stream("What is achatbot?"):
            print(chunk)

    asyncio.run(main())



import os
import cv2
import numpy as np
import torch
from typing import Union, List
import torch.nn.functional as F
from einops import rearrange
import shutil
import os.path as osp

from musetalk.models.vae import VAE
from musetalk.models.unet import UNet, PositionalEncoding


def load_all_model(
    vae_dir=os.path.join("models", "sd-vae"),
    unet_dir=os.path.join("models", "musetalkV15"),
    device=None,
):
    print(f"load vae model from {vae_dir}")
    vae = VAE(model_path=vae_dir)

    print(f"load unet model from {unet_dir}")
    unet_config = os.path.join(unet_dir, "musetalk.json")
    unet_model_path = os.path.join(unet_dir, "unet.pth")
    unet = UNet(unet_config=unet_config, model_path=unet_model_path, device=device)

    pe = PositionalEncoding(d_model=384)

    return vae, unet, pe


def get_file_type(video_path):
    _, ext = os.path.splitext(video_path)

    if ext.lower() in [".jpg", ".jpeg", ".png", ".bmp", ".tif", ".tiff"]:
        return "image"
    elif ext.lower() in [".avi", ".mp4", ".mov", ".flv", ".mkv"]:
        return "video"
    else:
        return "unsupported"


def get_video_fps(video_path):
    video = cv2.VideoCapture(video_path)
    fps = video.get(cv2.CAP_PROP_FPS)
    video.release()
    return fps


def datagen(
    whisper_chunks,
    vae_encode_latents,
    batch_size=8,
    delay_frame=0,
    device="cuda:0",
):
    whisper_batch, latent_batch = [], []
    for i, w in enumerate(whisper_chunks):
        idx = (i + delay_frame) % len(vae_encode_latents)
        latent = vae_encode_latents[idx]
        whisper_batch.append(w)
        latent_batch.append(latent)

        if len(latent_batch) >= batch_size:
            whisper_batch = torch.stack(whisper_batch)
            latent_batch = torch.cat(latent_batch, dim=0)
            yield whisper_batch, latent_batch
            whisper_batch, latent_batch = [], []

    # the last batch may smaller than batch size
    if len(latent_batch) > 0:
        whisper_batch = torch.stack(whisper_batch)
        latent_batch = torch.cat(latent_batch, dim=0)

        yield whisper_batch.to(device), latent_batch.to(device)


def cast_training_params(
    model: Union[torch.nn.Module, List[torch.nn.Module]],
    dtype=torch.float32,
):
    if not isinstance(model, list):
        model = [model]
    for m in model:
        for param in m.parameters():
            # only upcast trainable parameters into fp32
            if param.requires_grad:
                param.data = param.to(dtype)


def rand_log_normal(shape, loc=0.0, scale=1.0, device="cpu", dtype=torch.float32, generator=None):
    """Draws samples from an lognormal distribution."""
    rnd_normal = torch.randn(shape, device=device, dtype=dtype, generator=generator)  # N(0, I)
    sigma = (rnd_normal * scale + loc).exp()
    return sigma


def get_mouth_region(frames, image_pred, pixel_values_face_mask):
    # Initialize lists to store the results for each image in the batch
    mouth_real_list = []
    mouth_generated_list = []

    # Process each image in the batch
    for b in range(frames.shape[0]):
        # Find the non-zero area in the face mask
        non_zero_indices = torch.nonzero(pixel_values_face_mask[b])
        # If there are no non-zero indices, skip this image
        if non_zero_indices.numel() == 0:
            continue

        min_y, max_y = torch.min(non_zero_indices[:, 1]), torch.max(non_zero_indices[:, 1])
        min_x, max_x = torch.min(non_zero_indices[:, 2]), torch.max(non_zero_indices[:, 2])

        # Crop the frames and image_pred according to the non-zero area
        frames_cropped = frames[b, :, min_y:max_y, min_x:max_x]
        image_pred_cropped = image_pred[b, :, min_y:max_y, min_x:max_x]
        # Resize the cropped images to 256*256
        frames_resized = F.interpolate(
            frames_cropped.unsqueeze(0), size=(256, 256), mode="bilinear", align_corners=False
        )
        image_pred_resized = F.interpolate(
            image_pred_cropped.unsqueeze(0), size=(256, 256), mode="bilinear", align_corners=False
        )

        # Append the resized images to the result lists
        mouth_real_list.append(frames_resized)
        mouth_generated_list.append(image_pred_resized)

    # Convert the lists to tensors if they are not empty
    mouth_real = torch.cat(mouth_real_list, dim=0) if mouth_real_list else None
    mouth_generated = torch.cat(mouth_generated_list, dim=0) if mouth_generated_list else None

    return mouth_real, mouth_generated


def get_image_pred(pixel_values, ref_pixel_values, audio_prompts, vae, net, weight_dtype):
    with torch.no_grad():
        bsz, num_frames, c, h, w = pixel_values.shape

        masked_pixel_values = pixel_values.clone()
        masked_pixel_values[:, :, :, h // 2 :, :] = -1

        masked_frames = rearrange(masked_pixel_values, "b f c h w -> (b f) c h w")
        masked_latents = vae.encode(masked_frames).latent_dist.mode()
        masked_latents = masked_latents * vae.config.scaling_factor
        masked_latents = masked_latents.float()

        ref_frames = rearrange(ref_pixel_values, "b f c h w-> (b f) c h w")
        ref_latents = vae.encode(ref_frames).latent_dist.mode()
        ref_latents = ref_latents * vae.config.scaling_factor
        ref_latents = ref_latents.float()

        input_latents = torch.cat([masked_latents, ref_latents], dim=1)
        input_latents = input_latents.to(weight_dtype)
        timesteps = torch.tensor([0], device=input_latents.device)
        latents_pred = net(
            input_latents,
            timesteps,
            audio_prompts,
        )
        latents_pred = (1 / vae.config.scaling_factor) * latents_pred
        image_pred = vae.decode(latents_pred).sample
        image_pred = image_pred.float()

    return image_pred


def process_audio_features(cfg, batch, wav2vec, bsz, num_frames, weight_dtype):
    with torch.no_grad():
        audio_feature_length_per_frame = 2 * (
            cfg.data.audio_padding_length_left + cfg.data.audio_padding_length_right + 1
        )
        audio_feats = batch["audio_feature"].to(weight_dtype)
        audio_feats = wav2vec.encoder(audio_feats, output_hidden_states=True).hidden_states
        audio_feats = torch.stack(audio_feats, dim=2).to(weight_dtype)  # [B, T, 10, 5, 384]

        start_ts = batch["audio_offset"]
        step_ts = batch["audio_step"]
        audio_feats = torch.cat(
            [
                torch.zeros_like(audio_feats[:, : 2 * cfg.data.audio_padding_length_left]),
                audio_feats,
                torch.zeros_like(audio_feats[:, : 2 * cfg.data.audio_padding_length_right]),
            ],
            1,
        )
        audio_prompts = []
        for bb in range(bsz):
            audio_feats_list = []
            for f in range(num_frames):
                cur_t = (start_ts[bb] + f * step_ts[bb]) * 2
                audio_clip = audio_feats[
                    bb : bb + 1, cur_t : cur_t + audio_feature_length_per_frame
                ]

                audio_feats_list.append(audio_clip)
            audio_feats_list = torch.stack(audio_feats_list, 1)
            audio_prompts.append(audio_feats_list)
        audio_prompts = torch.cat(audio_prompts)  # B, T, 10, 5, 384
    return audio_prompts


def save_checkpoint(
    model, save_dir, ckpt_num, name="appearance_net", total_limit=None, logger=None
):
    save_path = os.path.join(save_dir, f"{name}-{ckpt_num}.pth")

    if total_limit is not None:
        checkpoints = os.listdir(save_dir)
        checkpoints = [d for d in checkpoints if d.endswith(".pth")]
        checkpoints = [d for d in checkpoints if name in d]
        checkpoints = sorted(checkpoints, key=lambda x: int(x.split("-")[1].split(".")[0]))

        if len(checkpoints) >= total_limit:
            num_to_remove = len(checkpoints) - total_limit + 1
            removing_checkpoints = checkpoints[0:num_to_remove]
            logger.info(
                f"{len(checkpoints)} checkpoints already exist, removing {len(removing_checkpoints)} checkpoints"
            )
            logger.info(f"removing checkpoints: {', '.join(removing_checkpoints)}")

            for removing_checkpoint in removing_checkpoints:
                removing_checkpoint = os.path.join(save_dir, removing_checkpoint)
                os.remove(removing_checkpoint)

    state_dict = model.state_dict()
    torch.save(state_dict, save_path)


def save_models(accelerator, net, save_dir, global_step, cfg, logger=None):
    unwarp_net = accelerator.unwrap_model(net)
    save_checkpoint(
        unwarp_net.unet,
        save_dir,
        global_step,
        name="unet",
        total_limit=cfg.total_limit,
        logger=logger,
    )


def delete_additional_ckpt(base_path, num_keep):
    dirs = []
    for d in os.listdir(base_path):
        if d.startswith("checkpoint-"):
            dirs.append(d)
    num_tot = len(dirs)
    if num_tot <= num_keep:
        return
    # ensure ckpt is sorted and delete the ealier!
    del_dirs = sorted(dirs, key=lambda x: int(x.split("-")[-1]))[: num_tot - num_keep]
    for d in del_dirs:
        path_to_dir = osp.join(base_path, d)
        if osp.exists(path_to_dir):
            shutil.rmtree(path_to_dir)


def seed_everything(seed):
    import random

    import numpy as np

    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed % (2**32))
    random.seed(seed)


def process_and_save_images(
    batch,
    image_pred,
    image_pred_infer,
    save_dir,
    global_step,
    accelerator,
    num_images_to_keep=10,
    syncnet_score=1,
):
    # Rearrange the tensors
    print("image_pred.shape: ", image_pred.shape)
    pixel_values_ref_img = rearrange(batch["pixel_values_ref_img"], "b f c h w -> (b f) c h w")
    pixel_values = rearrange(batch["pixel_values_vid"], "b f c h w -> (b f) c h w")

    # Create masked pixel values
    masked_pixel_values = batch["pixel_values_vid"].clone()
    _, _, _, h, _ = batch["pixel_values_vid"].shape
    masked_pixel_values[:, :, :, h // 2 :, :] = -1
    masked_pixel_values = rearrange(masked_pixel_values, "b f c h w -> (b f) c h w")

    # Keep only the specified number of images
    pixel_values = pixel_values[:num_images_to_keep, :, :, :]
    masked_pixel_values = masked_pixel_values[:num_images_to_keep, :, :, :]
    pixel_values_ref_img = pixel_values_ref_img[:num_images_to_keep, :, :, :]
    image_pred = image_pred.detach()[:num_images_to_keep, :, :, :]
    image_pred_infer = image_pred_infer.detach()[:num_images_to_keep, :, :, :]

    # Concatenate images
    concat = torch.cat(
        [
            masked_pixel_values * 0.5 + 0.5,
            pixel_values_ref_img * 0.5 + 0.5,
            image_pred * 0.5 + 0.5,
            pixel_values * 0.5 + 0.5,
            image_pred_infer * 0.5 + 0.5,
        ],
        dim=2,
    )
    print("concat.shape: ", concat.shape)

    # Create the save directory if it doesn't exist
    os.makedirs(f"{save_dir}/samples/", exist_ok=True)

    # Try to save the concatenated image
    try:
        # Concatenate images horizontally and convert to numpy array
        final_image = (
            torch.cat([concat[i] for i in range(concat.shape[0])], dim=-1)
            .permute(1, 2, 0)
            .cpu()
            .numpy()[:, :, [2, 1, 0]]
            * 255
        )
        # Save the image
        cv2.imwrite(
            f"{save_dir}/samples/sample_{global_step}_{accelerator.device}_SyncNetScore_{syncnet_score}.jpg",
            final_image,
        )
        print(
            f"Image saved successfully: {save_dir}/samples/sample_{global_step}_{accelerator.device}_SyncNetScore_{syncnet_score}.jpg"
        )
    except Exception as e:
        print(f"Failed to save image: {e}")



import numpy as np


def generate_silence(duration=3, sr=44100):
    """
    生成空白音频文件
    """
    silence = np.zeros(int(duration * sr))

    return silence


def generate_white_noise(duration=3, sr=44100):
    """
    生成白噪声
    duration: 持续时间(秒)
    sr: 采样率
    """
    # 生成随机噪声
    samples = np.random.normal(0, 1, size=int(duration * sr))
    # 归一化到 [-1, 1] 范围
    samples = samples / np.max(np.abs(samples))

    return samples


def generate_random_sine(duration=3, sr=44100):
    """
    生成随机频率的正弦波
    """
    # 随机选择频率 (20Hz-2000Hz)
    freq = np.random.uniform(20, 2000)

    # 生成时间序列
    t = np.linspace(0, duration, int(sr * duration))

    # 生成正弦波
    samples = np.sin(2 * np.pi * freq * t)

    return samples


def generate_random_notes(duration=3, sr=44100):
    """
    生成随机音符序列
    """
    # 定义一些音符频率 (C4-C5)
    notes = [261.63, 293.66, 329.63, 349.23, 392.00, 440.00, 493.88, 523.25]

    # 随机选择音符
    note_duration = 0.5  # 每个音符持续0.5秒
    num_notes = int(duration / note_duration)

    samples = np.array([])
    for _ in range(num_notes):
        freq = np.random.choice(notes)
        t = np.linspace(0, note_duration, int(sr * note_duration))
        note = np.sin(2 * np.pi * freq * t)
        # 添加淡入淡出效果
        note = note * np.hanning(len(note))
        samples = np.append(samples, note)

    # 归一化
    samples = samples / np.max(np.abs(samples))

    return samples


def generate_human_like_voice(duration=3, sr=44100):
    """
    生成类似人声的随机音频
    duration: 持续时间(秒)
    sr: 采样率
    """
    # 基础频率范围 (人声基频大约在85-255Hz之间)
    fundamental_freq = np.random.uniform(85, 255)

    # 生成时间序列
    t = np.linspace(0, duration, int(sr * duration))

    # 生成基础信号
    signal = np.zeros_like(t)

    # 添加基频和谐波
    for i in range(1, 6):
        # 每个谐波的振幅逐渐减小
        amplitude = 1.0 / i
        # 添加细微的频率变化模拟自然语音
        freq_variation = np.sin(2 * np.pi * 2 * t) * 10
        signal += amplitude * np.sin(2 * np.pi * (fundamental_freq * i + freq_variation) * t)

    # 添加音量包络
    num_syllables = int(duration * 2)  # 每秒约2个音节
    envelope = np.zeros_like(t)

    for i in range(num_syllables):
        center = i * (len(t) / num_syllables)
        width = len(t) / (num_syllables * 4)
        envelope += np.exp(-((t - center / sr) ** 2) / (2 * (width / sr) ** 2))

    envelope = envelope / np.max(envelope)
    signal = signal * envelope

    # 添加一些噪声模拟气流
    noise = np.random.normal(0, 0.01, len(signal))
    signal = signal + noise

    # 归一化
    signal = signal / np.max(np.abs(signal))

    # 应用低通滤波
    # signal = librosa.effects.preemphasis(signal, coef=0.95)

    return signal


"""
AUDIO_FUNCTION=generate_silence AUDIO_DURATION=3 AUDIO_SR=16000 python -m src.modules.speech.help.audio_mock 
AUDIO_FUNCTION=generate_white_noise AUDIO_DURATION=3 AUDIO_SR=16000 python -m src.modules.speech.help.audio_mock 
AUDIO_FUNCTION=generate_random_sine AUDIO_DURATION=3 AUDIO_SR=16000 python -m src.modules.speech.help.audio_mock 
AUDIO_FUNCTION=generate_random_notes AUDIO_DURATION=3 AUDIO_SR=16000 python -m src.modules.speech.help.audio_mock 
AUDIO_FUNCTION=generate_human_like_voice AUDIO_DURATION=3 AUDIO_SR=16000 python -m src.modules.speech.help.audio_mock 
"""

if __name__ == "__main__":
    import soundfile as sf
    import os
    from achatbot.common.types import ASSETS_DIR

    duration = int(os.getenv("AUDIO_DURATION", "3"))  # 默认3秒
    sr = int(os.getenv("AUDIO_SR", "44100"))  # 默认采样率44100Hz
    function_name = os.getenv("AUDIO_FUNCTION", "generate_silence")
    save_path = os.path.join(ASSETS_DIR, f"{function_name}.wav")

    # 获取对应的函数对象
    function = globals().get(function_name)
    if function is None:
        raise ValueError(f"未找到函数: {function_name}")

    samples = function(duration, sr)
    sf.write(save_path, samples, sr)
    info = sf.info(save_path)
    print(info)



import uuid
import logging
from pathlib import Path

from jinja2 import Template
from a2a.types import (
    Message,
    Part,
    Role,
    Task,
    TaskState,
    TextPart,
)
from google.adk import Agent
from google.adk.agents.callback_context import CallbackContext
from google.adk.agents.readonly_context import ReadonlyContext
from google.adk.models.lite_llm import LiteLlm
from google.adk.tools.tool_context import ToolContext

from src.services.a2a_multiagents import ADKBaseHostAgent
from . import convert_parts


class ADKSupervisorAgent(ADKBaseHostAgent):
    """The supervisor agent.

    This is the agent responsible for choosing which remote agents to send
    tasks to and coordinate/summary or directly return their work.

    like langchain/langgraph multi-agent supervisor agent:
    - https://github.com/weedge/doraemon-nb/blob/main/langchain/agent_supervisor.ipynb
    - https://github.com/weedge/doraemon-nb/blob/main/langchain/hierarchical_agent_teams.ipynb

    use http sse for streaming response from remote agents
    """

    def create_agent(self) -> Agent:
        tools = [] if len(self.remote_agent_connections) == 0 else [self.send_message]
        return Agent(
            model=LiteLlm(model=self.model),
            name=self.name or "supervisor_agent",
            instruction=self.root_instruction,
            before_model_callback=self.before_model_callback,
            description=(
                "This agent orchestrates the decomposition of the user request into"
                " tasks that can be performed by the child agents."
            ),
            tools=tools,
        )

    def root_instruction(self, context: ReadonlyContext) -> str:
        current_agent = self.check_state(context)
        prompt = (
            self.system_prompt
            or f"""You are an expert delegator that can delegate the user request to the
appropriate remote agents.

Execution:
- For actionable requests, you can use `send_message` to interact with remote agents to take action.

Be sure to include the remote agent name when you respond to the user.

Please rely on tools to address the request, and don't make up the response. If you are not sure, please ask the user for more details.
Focus on the most recent parts of the conversation primarily. It is not necessary to disclose which agent the information was obtained from.

Your output will be converted to audio, so please do not include special characters in your response.
Respond to what the user says in a creative and helpful way. 
Don't over-explain what you're doing. When making tool calls, respond with only short sentences.
"""
        )

        instruction = prompt
        dir_path = Path(__file__).parent
        with Path(dir_path / "tpl" / "supervisor.jinja").open("r") as f:
            instruction_template = Template(f.read())
            instruction = instruction_template.render(
                prompt=prompt, agents=self.agents, current_agent=current_agent["active_agent"]
            )
        logging.debug(f"{instruction=}")
        return instruction

    async def send_message(self, agent_name: str, message: str, tool_context: ToolContext):
        """Sends a task either streaming (if supported) or non-streaming.

        This will send a message to the remote agent named agent_name.

        Args:
          agent_name: The name of the agent to send the task to.
          message: The message to send to the agent for the task.
          tool_context: The tool context this method runs in.

        Yields:
          A dictionary of JSON data.
        """
        if agent_name not in self.remote_agent_connections:
            raise ValueError(f"Agent {agent_name} not found")
        state = tool_context.state
        state["agent"] = agent_name
        client = self.remote_agent_connections[agent_name]
        if not client:
            raise ValueError(f"Client not available for {agent_name}")
        task_id = state.get("task_id", None)
        context_id = state.get("context_id", None)
        message_id = state.get("message_id", None)
        task: Task
        if not message_id:
            message_id = str(uuid.uuid4())

        request_message = Message(
            role=Role.user,
            parts=[Part(root=TextPart(text=message))],
            message_id=message_id,
            context_id=context_id,
            task_id=task_id,
        )
        response = await client.send_message(request_message)
        if isinstance(response, Message):
            response = await convert_parts(response.parts, tool_context)
            logging.info(f"send_message response:{response}")
            return response
        task: Task = response
        # Assume completion unless a state returns that isn't complete
        state["session_active"] = task.status.state not in [
            TaskState.completed,
            TaskState.canceled,
            TaskState.failed,
            TaskState.unknown,
        ]
        if task.context_id:
            state["context_id"] = task.context_id
        state["task_id"] = task.id
        if task.status.state == TaskState.input_required:
            # Force user input back
            tool_context.actions.skip_summarization = True
            tool_context.actions.escalate = True
        elif task.status.state == TaskState.canceled:
            # Open question, should we return some info for cancellation instead
            raise ValueError(f"Agent {agent_name} task {task.id} is cancelled")
        elif task.status.state == TaskState.failed:
            # Raise error for failure
            raise ValueError(f"Agent {agent_name} task {task.id} failed")
        response = []
        if task.status.message:
            # Assume the information is in the task message.
            if ts := self.timestamp_extension.get_timestamp(task.status.message):
                response.append(f"[at {ts.astimezone().isoformat()}]")
            response.extend(await convert_parts(task.status.message.parts, tool_context))
        if task.artifacts:
            for artifact in task.artifacts:
                if ts := self.timestamp_extension.get_timestamp(artifact):
                    response.append(f"[at {ts.astimezone().isoformat()}]")
                response.extend(await convert_parts(artifact.parts, tool_context))
        logging.info(f"send_message response:{response}")
        return response

    def before_model_callback(self, callback_context: CallbackContext, llm_request):
        state = callback_context.state
        if "session_active" not in state or not state["session_active"]:
            state["session_active"] = True


"""
LOG_LEVEL=debug python -m src.services.a2a_multiagents.adk_supervisor_agent
python -m src.services.a2a_multiagents.adk_supervisor_agent
"""
if __name__ == "__main__":
    import os
    import asyncio
    import uuid

    from google.adk import Runner
    from google.adk.artifacts import InMemoryArtifactService
    from google.adk.memory.in_memory_memory_service import InMemoryMemoryService
    from google.adk.sessions.in_memory_session_service import InMemorySessionService
    from google.adk.agents.run_config import RunConfig, StreamingMode
    from google.genai import types

    from src.services.help.httpx import HTTPXClientWrapper
    from src.common.logger import Logger

    Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False, is_console=True)

    async def arun():
        http_client_wrapper = HTTPXClientWrapper()
        http_client_wrapper.start(timeout=60)
        host_agent = ADKSupervisorAgent(
            ["http://0.0.0.0:6666"], http_client_wrapper()
        )  # github repo describe is long-time task
        await host_agent.wait_init()  # sync init

        agent = host_agent.create_agent()

        session_service = InMemorySessionService()
        app_name = "A2A_app"
        user_id = "user1"
        host_runner = Runner(
            app_name=app_name,
            agent=agent,
            artifact_service=InMemoryArtifactService(),
            session_service=session_service,
            memory_service=InMemoryMemoryService(),
        )
        session = await session_service.create_session(app_name=app_name, user_id=user_id)
        async for event in host_runner.run_async(
            user_id=user_id,
            session_id=session.id,
            new_message=types.Content(
                parts=[types.Part.from_text(text="what's a chatbot?")], role="user"
            ),
            run_config=RunConfig(streaming_mode=StreamingMode.SSE),
        ):
            print("event-->", event)

        # await http_client_wrapper.stop()

    asyncio.run(arun())



import os
import numpy as np
import sys

def compute_wer(ref_file,
                hyp_file,
                cer_detail_file):
    rst = {
        'Wrd': 0,
        'Corr': 0,
        'Ins': 0,
        'Del': 0,
        'Sub': 0,
        'Snt': 0,
        'Err': 0.0,
        'S.Err': 0.0,
        'wrong_words': 0,
        'wrong_sentences': 0
    }

    hyp_dict = {}
    ref_dict = {}
    with open(hyp_file, 'r') as hyp_reader:
        for line in hyp_reader:
            key = line.strip().split()[0]
            value = line.strip().split()[1:]
            hyp_dict[key] = value
    with open(ref_file, 'r') as ref_reader:
        for line in ref_reader:
            key = line.strip().split()[0]
            value = line.strip().split()[1:]
            ref_dict[key] = value

    cer_detail_writer = open(cer_detail_file, 'w')
    for hyp_key in hyp_dict:
        if hyp_key in ref_dict:
           out_item = compute_wer_by_line(hyp_dict[hyp_key], ref_dict[hyp_key])
           rst['Wrd'] += out_item['nwords']
           rst['Corr'] += out_item['cor']
           rst['wrong_words'] += out_item['wrong']
           rst['Ins'] += out_item['ins']
           rst['Del'] += out_item['del']
           rst['Sub'] += out_item['sub']
           rst['Snt'] += 1
           if out_item['wrong'] > 0:
               rst['wrong_sentences'] += 1
           cer_detail_writer.write(hyp_key + print_cer_detail(out_item) + '\n')
           cer_detail_writer.write("ref:" + '\t' + "".join(ref_dict[hyp_key]) + '\n')
           cer_detail_writer.write("hyp:" + '\t' + "".join(hyp_dict[hyp_key]) + '\n')

    if rst['Wrd'] > 0:
        rst['Err'] = round(rst['wrong_words'] * 100 / rst['Wrd'], 2)
    if rst['Snt'] > 0:
        rst['S.Err'] = round(rst['wrong_sentences'] * 100 / rst['Snt'], 2)

    cer_detail_writer.write('\n')
    cer_detail_writer.write("%WER " + str(rst['Err']) + " [ " + str(rst['wrong_words'])+ " / " + str(rst['Wrd']) +
                            ", " + str(rst['Ins']) + " ins, " + str(rst['Del']) + " del, " + str(rst['Sub']) + " sub ]" + '\n')
    cer_detail_writer.write("%SER " + str(rst['S.Err']) + " [ " + str(rst['wrong_sentences']) + " / " + str(rst['Snt']) + " ]" + '\n')
    cer_detail_writer.write("Scored " + str(len(hyp_dict)) + " sentences, " + str(len(hyp_dict) - rst['Snt']) + " not present in hyp." + '\n')

     
def compute_wer_by_line(hyp,
                        ref):
    hyp = list(map(lambda x: x.lower(), hyp))
    ref = list(map(lambda x: x.lower(), ref))

    len_hyp = len(hyp)
    len_ref = len(ref)

    cost_matrix = np.zeros((len_hyp + 1, len_ref + 1), dtype=np.int16)

    ops_matrix = np.zeros((len_hyp + 1, len_ref + 1), dtype=np.int8)

    for i in range(len_hyp + 1):
        cost_matrix[i][0] = i
    for j in range(len_ref + 1):
        cost_matrix[0][j] = j

    for i in range(1, len_hyp + 1):
        for j in range(1, len_ref + 1):
            if hyp[i - 1] == ref[j - 1]:
                cost_matrix[i][j] = cost_matrix[i - 1][j - 1]
            else:
                substitution = cost_matrix[i - 1][j - 1] + 1
                insertion = cost_matrix[i - 1][j] + 1
                deletion = cost_matrix[i][j - 1] + 1

                compare_val = [substitution, insertion, deletion]

                min_val = min(compare_val)
                operation_idx = compare_val.index(min_val) + 1
                cost_matrix[i][j] = min_val
                ops_matrix[i][j] = operation_idx

    match_idx = []
    i = len_hyp
    j = len_ref
    rst = {
        'nwords': len_ref,
        'cor': 0,
        'wrong': 0,
        'ins': 0,
        'del': 0,
        'sub': 0
    }
    while i >= 0 or j >= 0:
        i_idx = max(0, i)
        j_idx = max(0, j)

        if ops_matrix[i_idx][j_idx] == 0:  # correct
            if i - 1 >= 0 and j - 1 >= 0:
                match_idx.append((j - 1, i - 1))
                rst['cor'] += 1

            i -= 1
            j -= 1

        elif ops_matrix[i_idx][j_idx] == 2:  # insert
            i -= 1
            rst['ins'] += 1

        elif ops_matrix[i_idx][j_idx] == 3:  # delete
            j -= 1
            rst['del'] += 1

        elif ops_matrix[i_idx][j_idx] == 1:  # substitute
            i -= 1
            j -= 1
            rst['sub'] += 1

        if i < 0 and j >= 0:
            rst['del'] += 1
        elif j < 0 and i >= 0:
            rst['ins'] += 1

    match_idx.reverse()
    wrong_cnt = cost_matrix[len_hyp][len_ref]
    rst['wrong'] = wrong_cnt

    return rst

def print_cer_detail(rst):
    return ("(" + "nwords=" + str(rst['nwords']) + ",cor=" + str(rst['cor'])
            + ",ins=" + str(rst['ins']) + ",del=" + str(rst['del']) + ",sub="
            + str(rst['sub']) + ") corr:" + '{:.2%}'.format(rst['cor']/rst['nwords'])
            + ",cer:" + '{:.2%}'.format(rst['wrong']/rst['nwords']))

if __name__ == '__main__':
    if len(sys.argv) != 4:
        print("usage : python compute-wer.py test.ref test.hyp test.wer")
        sys.exit(0)

    ref_file = sys.argv[1]
    hyp_file = sys.argv[2]
    cer_detail_file = sys.argv[3]
    compute_wer(ref_file, hyp_file, cer_detail_file)
